"""send -- Construct and send Apple events. """

from .ae import newappleevent, stringsforosstatus, MacOSError
from . import kae

from .aemcodecs import Codecs

__all__ = ['Event', 'EventError']

######################################################################
# PRIVATE
######################################################################

_defaultcodecs = Codecs()

def sendappleevent(evt, flags, timeout):
	""" Default function for sending Apple events.
		evt : aem.ae.AEDesc -- the AppleEvent to send
		flags : int -- send mode flags
		timeout : int -- timeout delay
		Result : aem.ae.AEDesc -- the reply AppleEvent
	"""
	return evt.send(flags, timeout)

######################################################################
# PUBLIC
######################################################################


class Event:
	"""Represents an Apple event (serialised message)."""
	
	def __init__(self, address, event, params={}, atts={},
			transaction=kae.kAnyTransactionID, returnid= kae.kAutoGenerateReturnID, 
			codecs=_defaultcodecs, createproc=newappleevent, sendproc=sendappleevent):
		"""Called by aem.Application.event(); users shouldn't instantiate this class themselves.
			address : AEAddressDesc -- the target application
			event : bytes -- 8-letter code indicating event's class and id, e.g. b'coregetd'
			params : dict -- a dict of form {AE_code:anything,...} containing zero or more event parameters (message arguments)
			atts : dict -- a dict of form {AE_code:anything,...} containing zero or more event attributes (event info)
			transaction : int -- transaction number (default = kAnyTransactionID)
			returnid : int  -- reply event's ID (default = kAutoGenerateReturnID)
			codecs : Codecs -- user can provide custom parameter & result encoder/decoder (default = standard codecs); supplied by Application class
			createproc : function -- function to create a new AppleEvent descriptor
			sendproc : function -- function to send an AppleEvent descriptor
		"""
		self._eventcode = event
		self._codecs = codecs
		self._sendproc = sendproc
		self.AEM_event = createproc(event[:4], event[4:], address, returnid, transaction)
		for key, value in atts.items():
			self.AEM_event.setattr(key, codecs.pack(value))
		for key, value in params.items():
			self.AEM_event.setparam(key, codecs.pack(value))
	
	# Public
	
	def send(self, timeout= kae.kAEDefaultTimeout, flags= kae.kAECanSwitchLayer + kae.kAEWaitReply):
		"""Send this Apple event (may be called any number of times).
			timeout : int | aem.k.DefaultTimeout | aem.k.NoTimeout -- number of ticks to wait for target process
					to reply before raising timeout error (default=DefaultTimeout)
			flags : int -- bitwise flags [1] indicating how target process should handle event (default=WaitReply)
			Result : anything -- value returned by application, if any
			
			[1] aem.k provides the following constants for convenience:
			
				[ aem.k.NoReply | aem.k.QueueReply | aem.k.WaitReply ]
				[ aem.k.DontReconnect ]
				[ aem.k.WantReceipt ]
				[ aem.k.NeverInteract | aem.k.CanInteract | aem.k.AlwaysInteract ]
				[ aem.k.CanSwitchLayer ]
		"""
		try:
			replyevent = self._sendproc(self.AEM_event, flags, timeout)
		except MacOSError as err: # an OS-level error occurred
			if not (self._eventcode == b'aevtquit' and err.args[0] == -609): # Ignore invalid connection error (-609) when quitting
				raise EventError(err.args[0]) from err
		else: # decode application's reply, if any
			if replyevent.type != kae.typeNull:
				eventresult = dict([replyevent.getitem(i + 1, kae.typeWildCard) 
						for i in range(replyevent.count())])
				# note: while Apple docs say that both keyErrorNumber and keyErrorString should be
				# tested for when determining if an error has occurred, AppleScript tests for keyErrorNumber
				# only, so do the same here for compatibility
				if kae.keyErrorNumber in eventresult: # an application-level error occurred
					# note: uses standard codecs to unpack error info to ensure consistent conversion
					errornum = _defaultcodecs.unpack(eventresult[kae.keyErrorNumber])
					if errornum != 0: # Stupid Finder returns non-error error number and message for successful move/duplicate command, so just ignore it
						errormsg = eventresult.get(kae.keyErrorString)
						if errormsg:
							errormsg = _defaultcodecs.unpack(errormsg)
						raise EventError(errornum, errormsg, eventresult)
				if kae.keyAEResult in eventresult: # application has returned a value
					# note: unpack result with [optionally] user-specified codecs, allowing clients to customise unpacking (e.g. appscript)
					return self._codecs.unpack(eventresult[kae.keyAEResult])



######################################################################


class EventError(MacOSError):
	""" Raised by aem.Event.send() when sending an event fails; contains error information 
		provided by Apple Event Manager or target application.
		
		Notes:
			
			- the 'raw' attribute contains either a dict containing the reply event's 
				raw parameters, or an empty dict if the error occurred while sending 
				the outgoing event; used by appscript.CommandError; third-parties 
				should avoid using it directly
	"""
	
	_carbonerrors = { # Following error descriptions are mostly cribbed from AppleScript Language Guide.
		# OS errors
		-34: "Disk is full.",
		-35: "Disk wasn't found.",
		-37: "Bad name for file.",
		-38: "File wasn't open.",
		-39: "End of file error.",
		-42: "Too many files open.",
		-43: "File wasn't found.",
		-44: "Disk is write protected.",
		-45: "File is locked.",
		-46: "Disk is locked.",
		-47: "File is busy.",
		-48: "Duplicate file name.",
		-49: "File is already open.",
		-50: "Parameter error.",
		-51: "File reference number error.",
		-61: "File not open with write permission.",
		-108: "Out of memory.",
		-120: "Folder wasn't found.",
		-124: "Disk is disconnected.",
		-128: "User canceled.",
		-192: "A resource wasn't found.",
		-600: "Application isn't running.",
		-601: "Not enough room to launch application with special requirements.",
		-602: "Application is not 32-bit clean.",
		-605: "More memory is needed than is specified in the size resource.",
		-606: "Application is background-only.",
		-607: "Buffer is too small.",
		-608: "No outstanding high-level event.",
		-609: "Connection is invalid.",
		-904: "Not enough system memory to connect to remote application.",
		-905: "Remote access is not allowed.",
		-906: "Application isn't running or program linking isn't enabled.",
		-915: "Can't find remote machine.",
		-30720: "Invalid date and time.",
		# AE errors
		-1700: "Can't make some data into the expected type.",
		-1701: "Some parameter is missing for command.",
		-1702: "Some data could not be read.",
		-1703: "Some data was the wrong type.",
		-1704: "Some parameter was invalid.",
		-1705: "Operation involving a list item failed.",
		-1706: "Need a newer version of the Apple Event Manager.",
		-1707: "Event isn't an Apple event.",
		-1708: "Application could not handle this command.",
		-1709: "AEResetTimer was passed an invalid reply.",
		-1710: "Invalid sending mode was passed.",
		-1711: "User canceled out of wait loop for reply or receipt.",
		-1712: "Apple event timed out.",
		-1713: "No user interaction allowed.",
		-1714: "Wrong keyword for a special function.",
		-1715: "Some parameter wasn't understood.",
		-1716: "Unknown Apple event address type.",
		-1717: "The handler is not defined.",
		-1718: "Reply has not yet arrived.",
		-1719: "Can't get reference. Invalid index.",
		-1720: "Invalid range.",
		-1721: "Wrong number of parameters for command.",
		-1723: "Can't get reference. Access not allowed.",
		-1725: "Illegal logical operator called.",
		-1726: "Illegal comparison or logical.",
		-1727: "Expected a reference.",
		-1728: "Can't get reference.",
		-1729: "Object counting procedure returned a negative count.",
		-1730: "Container specified was an empty list.",
		-1731: "Unknown object type.",
		-1739: "Attempting to perform an invalid operation on a null descriptor.",
		-1743: "The user has declined permission.",
		-1744: "User consent is required, but the user has not yet been prompted for it.",
		# Application scripting errors
		-10000: "Apple event handler failed.",
		-10001: "Type error.",
		-10002: "Invalid key form.",
		-10003: "Can't set reference to given value. Access not allowed.",
		-10004: "A privilege violation occurred.",
		-10005: "The read operation wasn't allowed.",
		-10006: "Can't set reference to given value.",
		-10007: "The index of the event is too large to be valid.",
		-10008: "The specified object is a property, not an element.",
		-10009: "Can't supply the requested descriptor type for the data.",
		-10010: "The Apple event handler can't handle objects of this class.",
		-10011: "Couldn't handle this command because it wasn't part of the current transaction.",
		-10012: "The transaction to which this command belonged isn't a valid transaction.",
		-10013: "There is no user selection.",
		-10014: "Handler only handles single objects.",
		-10015: "Can't undo the previous Apple event or user action.",
		-10023: "Enumerated value is not allowed for this property.",
		-10024: "Class can't be an element of container.",
		-10025: "Illegal combination of properties settings.",
	}
	
	# Following Cocoa Scripting error descriptions taken from:
	# http://developer.apple.com/documentation/Cocoa/Reference/Foundation/ObjC_classic/Classes/NSScriptCommand.html
	# http://developer.apple.com/documentation/Cocoa/Reference/Foundation/ObjC_classic/Classes/NSScriptObjectSpecifier.html

	_cocoaerrors = (
		('NSReceiverEvaluationScriptError', 'The object or objects specified by the direct parameter to a command could not be found.'),
		('NSKeySpecifierEvaluationScriptError', 'The object or objects specified by a key (for commands that support key specifiers) could not be found.'),
		('NSArgumentEvaluationScriptError', 'The object specified by an argument could not be found.'),
		('NSReceiversCantHandleCommandScriptError', "The receivers don't support the command sent to them."),
		('NSRequiredArgumentsMissingScriptError', 'An argument (or more than one argument) is missing.'),
		('NSArgumentsWrongScriptError', 'An argument (or more than one argument) is of the wrong type or is otherwise invalid.'),
		('NSUnknownKeyScriptError', 'An unidentified error occurred; indicates an error in the scripting support of your application.'),
		('NSInternalScriptError', 'An unidentified internal error occurred; indicates an error in the scripting support of your application.'),
		('NSOperationNotSupportedForKeyScriptError', 'The implementation of a scripting command signaled an error.'),
		('NSCannotCreateScriptCommandError', 'Could not create the script command; an invalid or unrecognized Apple event was received.'),
		('NSNoSpecifierError', 'No error encountered.'),
		('NSNoTopLevelContainersSpecifierError', 'Someone called evaluate with nil.'),
		('NSContainerSpecifierError', 'Error evaluating container specifier.'),
		('NSUnknownKeySpecifierError', 'Receivers do not understand the key.'),
		('NSInvalidIndexSpecifierError', 'Index out of bounds.'),
		('NSInternalSpecifierError', 'Other internal error.'),
		('NSOperationNotSupportedForKeySpecifierError', 'Attempt made to perform an unsupported operation on some key.'),
	)
	
	def __init__(self, number, message=None, raw=None):
		MacOSError.__init__(self, number)
		self._number, self._message, self._raw = number, str(message or ''), raw
	
	raw = property(lambda self: self._raw or {}, 
			doc="dict -- raw error data from reply event, if any (note: clients should not need to use this directly)")
	
	def __repr__(self):
		return "aem.EventError({!r}, {!r}, {!r})".format(self._number, self._message, self._raw)
		
	def __int__(self):
		return self._number
	
	def __str__(self):
		return "Command failed: {} ({})".format(self.errormessage, self.errornumber)
	
	# basic error info (an error number is always given by AEM/application;
	# message is either supplied by application or generated here)	
	errornumber = property(lambda self: self._number, doc="int -- Mac OS error number")
	
	def errormessage(self):
		message = self._message
		if self._number > 0 and message:
			for name, description in self._cocoaerrors:
				if message.startswith(name):
					message = '{} ({})'.format(message, description)
					break
		elif not message:
			message = self._carbonerrors.get(self._number)
			if not message:
				message = stringsforosstatus(self._number)[1] or 'OS error'
		return message
	errormessage = property(errormessage, 
			doc="str -- application-supplied/generic error description")
	
	# extended error info (some apps may return additional error info, though most don't)

	def _errorinfo(self, key):
		if self._raw:
			desc = self._raw.get(key)
			if desc:
				return _defaultcodecs.unpack(desc)
		return None
	
	offendingobject = property(lambda self: self._errorinfo(kae.kOSAErrorOffendingObject),
			doc="anything | None -- object that caused the error, if given by application")
	expectedtype = property(lambda self: self._errorinfo(kae.kOSAErrorExpectedType),
			doc="anything | None -- object that caused a coercion error, if given by application")
	partialresult = property(lambda self: self._errorinfo(kae.kOSAErrorPartialResult),
			doc="anything | None -- part of return value constructed before error occurred, if given by application")




"""typewrappers -- wrapper classes for AE type and enumeration codes """

######################################################################
# PUBLIC
######################################################################

class AETypeBase: # base class; exposed for typechecking purposes
	
	def __init__(self, code):
		# Check arg is a 4-byte code (while ae.newdesc() verifies descriptor type codes okay, it doesn't verify data size so wouldn't catch bad values at packing time):
		if not isinstance(code, bytes):
			raise TypeError('invalid code (not a bytes object): {!r}'.format(code))
		elif len(code) != 4:
			raise ValueError('invalid code (not four bytes long): {!r}'.format(code))
		self._code = code
	
	code = property(lambda self:self._code)
	
	def __hash__(self): 
		return hash(self._code)
	
	def __eq__(self, val):
		return val.__class__ == self.__class__ and val.code == self._code
	
	def __ne__(self, val):
		return not self == val
		
	def __repr__(self):
		return "aem.{}({})".format(self.__class__.__name__, self._code)


class AEType(AETypeBase):
	"""An AE type."""


class AEEnum(AETypeBase):
	"""An AE enumeration."""


class AEProp(AETypeBase):
	"""An AE property code."""


class AEKey(AETypeBase):
	"""An AE keyword."""




"""aem -- Mid-level wrapper for building and sending Apple events. """

import struct

from . import ae, kae, findapp, mactypes, aemconnect
from .aemconnect import CantLaunchApplicationError
from .aemsend import Event, EventError, newappleevent, sendappleevent
from .aemcodecs import Codecs
from .aemreference import app, con, its, customroot, Query
from .typewrappers import AETypeBase, AEType, AEEnum, AEProp, AEKey


__all__ = [
	'ae', 'kae',
	'findapp', 'mactypes',
	'Application',
	'Event', 'EventError',
	'CantLaunchApplicationError',
	'Codecs', 
	'app', 'con', 'its', 'customroot', 'Query',
	'AETypeBase', 'AEType', 'AEEnum', 'AEProp', 'AEKey',
]


######################################################################
# PRIVATE
######################################################################

_defaultcodecs = Codecs()

######################################################################
# PUBLIC
######################################################################

class Application(Query):
	"""Target application for Apple events."""
	
	# aem.Application subclasses can override these attributes to modify the creation
	# and sending of AppleEvent descriptors
	_createproc = staticmethod(newappleevent)
	_sendproc = staticmethod(sendappleevent)
	
	# aem.Application subclasses can override this attribute (typically with a subclass 
	# of the standard aemsend.Event class) to have the event() method return a different
	# class instance; simpler than overriding the event() method
	_Event = Event

	# need to keep a local copy of this constant to avoid upsetting Application.__del__() 
	# at cleanup time, otherwise it may be disposed of before __del__() can use it
	_transaction = _kAnyTransactionID = kae.kAnyTransactionID
	
	def __init__(self, path=None, pid=None, url=None, desc=None, 
			codecs= _defaultcodecs, newinstance=False, hide=False):
		"""
			path : str | None -- full path to local application
			pid : int | None -- Unix process id for local process
			url : str | None -- url for remote process
			desc : AEAddressDesc | None -- AEAddressDesc for application
			codecs : Codecs -- used to convert Python values to AEDescs and vice-versa
			newinstance : bool -- launch a new application instance?
			hide : bool -- hide after launch?
			
			Notes: 
			
				- If no path, pid, url or aedesc is given, target will be 'current application'.
				
				- If path is given, application will be launched automatically; if pid, url or 
					desc is given, user is responsible for ensuring application is running 
					before sending it any events.
				
				- The newinstance and hide options only apply when specifying application
					by path.
		"""
		self._path, self._codecs, self._newinstance, self._hide = path, codecs, newinstance, hide
		if path:
			self._address = aemconnect.localapp(path, newinstance, hide)
			self.AEM_identity = ('path', path)
		elif pid:
			self._address = aemconnect.localappbypid(pid)
			self.AEM_identity = ('pid', pid)
		elif url:
			self._address = aemconnect.remoteapp(url)
			self.AEM_identity = ('url', url)
		elif desc:
			self._address = desc
			self.AEM_identity = ('desc', (desc.type, desc.data))
		else:
			self._address = aemconnect.currentapp
			self.AEM_identity = ('current', None)
	
	def __repr__(self):
		args = []
		if self.AEM_identity[0] == 'desc':
			args.append('desc={!r}'.format(self._address))
		elif self.AEM_identity[0] == 'path':
			args.append(repr(self.AEM_identity[1]))
		elif self.AEM_identity[0] != 'current':
			args.append('{}={!r}'.format(*self.AEM_identity))
		if self._codecs != _defaultcodecs:
			args.append('codecs={!r}'.format(self._codecs))
		if self._newinstance:
			args.append('newinstance={!r}'.format(struct.unpack('II', self._address.data)))
		if self._hide:
			args.append('hide=True')
		modulename = '{}.'.format(self.__class__.__module__)
		if modulename == 'aem.send.':
			modulename = 'aem.'
		elif modulename == '__main__.':
			modulename = ''
		return '{}{}({})'.format(modulename, self.__class__.__name__, ', '.join(args))
			
	__str__ = __repr__
	
	def __eq__(self, val):
		return self is val or (
				self.__class__ == val.__class__ and 
				self.AEM_identity == val.AEM_identity)
	
	def __ne__(self, val):
		return not self == val
	
	def __hash__(self):
		return hash(self.AEM_identity)
	
	def AEM_comparable(self):
		return ['AEMApplication', self.AEM_identity]
	
	def AEM_packself(self, codecs):
		return self._address
	
	def __del__(self):
		# If user forgot to close a transaction before throwing away the Application object 
		# that opened it, try to close it for them. Otherwise application will be left in 
		# mid-transaction, preventing anyone else from using it.
		if self._transaction != self._kAnyTransactionID:
			self.endtransaction()
	
	#######
	# Utility functions; placed here for convenience
	
	# Launch a local application without sending it the usual 'run' event (aevtoapp):
	launch = staticmethod(aemconnect.launchapp)
	
	# Check if an application specified by path/pid/URL/AEAddressDesc is running:
	processexistsforpath = staticmethod(aemconnect.processexistsforpath)
	processexistsforpid = staticmethod(aemconnect.processexistsforpid)
	processexistsforurl = staticmethod(aemconnect.processexistsforurl)
	processexistsfordesc = staticmethod(aemconnect.processexistsfordesc)
	
	#######
	
	addressdesc = property(lambda self: self._address)
	
	def reconnect(self):
		"""If application has quit since this Application object was created, its 
			AEAddressDesc is no longer valid so this Application object 
			will not work even when application is restarted. reconnect() will 
			update this Application object's AEAddressDesc so it's valid again.
		
			Notes:
			
			- This only works for Application objects specified by path, not by
				URL or AEDesc. Also, any Event objects created prior to calling 
				reconnect() will still be invalid.
			
			- If the Application object was created with newinstance=True, calling
				reconnect() will launch a new application instance and connect 
				to that each time it is called. Otherwise it will reconnect to the
				first existing application instance it finds, and only launches a new
				instance if none are found.
		"""
		if self._path:
			self._address = aemconnect.localapp(self._path, self._newinstance, self._hide)
	
	def permissiontoautomate(self, eventclass=kae.typeWildCard, eventid=kae.typeWildCard, askuserifneeded=False):
		"""Throws if current process is not authorized to send an Apple event to the target application.
		
			Errors include:
	
			- errAEEventNotPermitted (-1743): the user has declined permission.
		
			- errAEEventWouldRequireUserConsent (-1744): user consent is required for this, but the user has not yet been prompted for it. You need to pass False for askUserIfNeeded to get this.
		
			- procNotFound (-600): the specified app is not currently running.
			
			Throws NotImplementedError on 10.13 and earlier. (Permission is always available.)
			
			## Known Issues ##
			
			AEDeterminePermissionToAutomateTarget() is buggy on 10.14. See rdar://44049802 (fixed in 10.14.3)
		"""
		self._address.permissiontoautomate(eventclass, eventid, askuserifneeded)
	
	##
	
	def event(self, event, params={}, atts={}, returnid=kae.kAutoGenerateReturnID, codecs=None):
		"""Construct an Apple event.
			event  : str -- 8-letter code indicating event's class, e.g. 'coregetd'
			params : dict -- a dict of form {AE_code:anything,...} containing zero or more 
					event parameters (message arguments)
			atts : dict -- a dict of form {AE_code:anything,...} containing zero or more 
					event attributes (event info)
			returnid : int  -- reply event's ID (default = kAutoGenerateReturnID)
			codecs : Codecs | None -- custom codecs to use when packing/unpacking this
					event; if None, codecs supplied in Application constructor are used
		"""
		return self._Event(self._address, event, params, atts, self._transaction, returnid, 
				codecs or self._codecs, self._createproc, self._sendproc)
	
	def begintransaction(self, session=None):
		"""Begin a transaction."""
		if self._transaction != self._kAnyTransactionID:
			raise RuntimeError("Transaction is already active.")
		self._transaction = self._Event(self._address, 'miscbegi', 
				session is not None and {'----':session} or {}, codecs=_defaultcodecs).send()
	
	def aborttransaction(self):
		"""Abort the current transaction."""
		if self._transaction == self._kAnyTransactionID:
			raise RuntimeError("No transaction is active.")
		self._Event(self._address, 'miscttrm', transaction=self._transaction).send()
		self._transaction = self._kAnyTransactionID
	
	def endtransaction(self):
		"""End the current transaction."""
		if self._transaction == self._kAnyTransactionID:
			raise RuntimeError("No transaction is active.")
		self._Event(self._address, 'miscendt', transaction=self._transaction).send()
		self._transaction = self._kAnyTransactionID




"""referencerenderer -- Generates string representations of appscript references from aem object specifiers. """

import struct
from .terminology import kProperty, kElement
from aem import Query

######################################################################
# PRIVATE
######################################################################

_property_ = property

class _Formatter:
	def __init__(self, appdata, nested=False):
		self._appdata = appdata
		if nested:
			self.root = 'app'
		elif self._appdata.constructor == 'current':
			self.root = 'app()'
		elif self._appdata.constructor == 'path':
			argstr = ''
			newinstancearg = self._appdata.aemconstructoroptions.get('newinstance')
			if newinstancearg:
				argstr += ', newinstance={!r}'.format(
						struct.unpack('II', self._appdata.target().addressdesc.data) if 
								self._appdata.isconnected else True)
			self.root = 'app({!r}{})'.format(self._appdata.identifier, argstr)
		else:
			self.root = 'app({}={!r})'.format(self._appdata.constructor, self._appdata.identifier)
		self.result = ''
	
	def _format(self, val):
		if isinstance(val, Query):
			return renderreference(self._appdata, val, True)
		else:
			return repr(val)
	
	# reference roots
	
	def app(self):
		self.result += self.root
		return self
	app = _property_(app)
	
	def con(self):
		self.result += 'con'
		return self
	con = _property_(con)
	
	def its(self):
		self.result += 'its'
		return self
	its = _property_(its)

	# insertion locs
	
	def beginning(self):
		self.result += '.beginning'
		return self
	beginning = _property_(beginning)
	
	def end(self):
		self.result += '.end'
		return self
	end = _property_(end)
	
	def before(self):
		self.result += '.before'
		return self
	before = _property_(before)
	
	def after(self):
		self.result += '.after'
		return self
	after = _property_(after)
	
	# property, elements specifiers
	
	def property(self, code):
		try:
			self.result += '.' + self._appdata.referencebycode()[kProperty+code][1]
		except KeyError:
			self.result += '.' + self._appdata.referencebycode()[kElement+code][1]
		return self

	def elements(self, code):
		try:
			self.result += '.' + self._appdata.referencebycode()[kElement+code][1]
		except KeyError:
			self.result += '.' + self._appdata.referencebycode()[kProperty+code][1]
		return self
	
	# single-element selectors
	
	def first(self):
		self.result += '.first'
		return self
	first = _property_(first)
	
	def middle(self):
		self.result += '.middle'
		return self
	middle = _property_(middle)
	
	def last(self):
		self.result += '.last'
		return self
	last = _property_(last)
	
	def any(self):
		self.result += '.any'
		return self
	any = _property_(any)
	
	def byindex(self, sel):
		self.result += '[{!r}]'.format(sel)
		return self
	
	byname = byindex
	
	def byid(self, sel):
		self.result += '.ID({!r})'.format(sel)
		return self
	
	def previous(self, sel):
		self.result += '.previous({!r})'.format(self._appdata.typebycode()[sel])
		return self
	
	def next(self, sel):
		self.result += '.next({!r})'.format(self._appdata.typebycode()[sel])
		return self
	
	# multi-element selectors
	
	def byrange(self, sel1, sel2):
		self.result += '[{}:{}]'.format(self._format(sel1), self._format(sel2))
		return self
		
	def byfilter(self, sel):
		self.result += '[{}]'.format(self._format(sel))
		return self
	
	# comparison tests
	
	def gt(self, sel):
		self.result += ' > {}'.format(self._format(sel))
		return self
	
	def ge(self, sel):
		self.result += ' >= {}'.format(self._format(sel))
		return self
	
	def eq(self, sel):
		self.result += ' == {}'.format(self._format(sel))
		return self
	
	def ne(self, sel):
		self.result += ' != {}'.format(self._format(sel))
		return self
	
	def lt(self, sel):
		self.result += ' < {}'.format(self._format(sel))
		return self
	
	def le(self, sel):
		self.result += ' <= {}'.format(self._format(sel))
		return self
	
	def beginswith(self, sel):
		self.result += '.beginswith({})'.format(self._format(sel))
		return self
	
	def endswith(self, sel):
		self.result += '.endswith({})'.format(self._format(sel))
		return self
	
	def contains(self, sel):
		self.result += '.contains({})'.format(self._format(sel))
		return self
	
	def isin(self, sel):
		self.result += '.isin({})'.format(self._format(sel))
		return self
	
	# logical tests
	
	def AND(self, *operands):
		self.result = '({}).AND({})'.format(self.result, ', '.join([self._format(o) for o in operands]))
		return self
		
	def OR(self, *operands):
		self.result = '({}).OR({})'.format(self.result, ', '.join([self._format(o) for o in operands]))
		return self
	
	def NOT(self):
		self.result = '({}).NOT'.format(self.result)
		return self
	NOT = _property_(NOT)


######################################################################
# PUBLIC
######################################################################

def renderreference(appdata, aemreference, nested=False):
	"""Take an aem reference, e.g.:
	
		app.elements('docu').byindex(1).property('ctxt')
	
	and an AppData instance containing application's location and terminology, and render an appscript-style reference, e.g.:
	
		"app('/Applications/TextEdit.app').documents[1].text"
		
	Used by Reference.__repr__().
	"""
	f = _Formatter(appdata, nested)
	try:
		aemreference.AEM_resolve(f)
	except:
		return '{}.AS_newreference({!r})'.format(f.root, aemreference)
	return f.result
	



"""terminologyparser -- Basic aete parser to construct name-code terminology tables from an application's class, enumerator, property, element and command definitions. 

The tables returned by this module are an intermediate format, suitable for exporting to Python modules via terminology.dump. The terminology module will convert these intermediate tables into the final format used in AppData objects. """

from struct import pack, unpack
import string

from aem import ae, kae
from .reservedkeywords import kReservedKeywords


######################################################################
# PRIVATE
######################################################################

class Parser:
	_keywordcache = {}
	_reservedwords = set(kReservedKeywords)
	_specialconversions = {
			' ': '_',
			'-': '_',
			'&': 'and',
			'/': '_',
	}
	_legalchars = string.ascii_letters + '_'
	_alphanum = _legalchars + string.digits
		
	def __init__(self):
		# terminology tables; order is significant where synonym definitions occur
		self.commands = {}
		self.properties = []
		self.elements = []
		self.classes = []
		self.enumerators = []
		# use sets to record previously found definitions, and avoid adding duplicates to lists
		# (i.e. 'name+code not in <set>' is quicker than using 'name+code not in <list>')
		self._foundproperties = set()
		self._foundelements = set()
		self._foundclasses = set()
		self._foundenumerators = set()
		# ideally, aetes should define both singular and plural names for each class, but
		# some define only one or the other so we need to fill in any missing ones afterwards
		self._spareclassnames = {}
		self._foundclasscodes = set()
		self._foundelementcodes = set()
		
	def integer(self):
		"""Read a 2-byte integer."""
		self._ptr += 2
		return unpack("H", self._data[self._ptr - 2:self._ptr])[0]
	
	def word(self):
		"""Read a 4-byte OSType."""
		self._ptr += 4
		return self._data[self._ptr - 4:self._ptr] # big-endian
	
	def name(self):
		"""Read a MacRoman-encoded Pascal keyword string."""
		count = self._data[self._ptr]
		self._ptr += 1 + count
		s = self._data[self._ptr - count:self._ptr].decode('macroman')
		if s not in self._keywordcache:
			legal = self._legalchars
			res = ''
			for c in s:
				if c in legal:
					res += c
				elif c in self._specialconversions:
					res += self._specialconversions[c]
				else:
					if res == '':
						res = '_' # avoid creating an invalid identifier
					res += '0x{:X}'.format(ord(c))
				legal = self._alphanum
			if res in self._reservedwords or res.startswith('_') or res.startswith('AS_'):
				res += '_'
			self._keywordcache[s] = str(res)
		return self._keywordcache[s]

	
	##
	
	def parsecommand(self):
		name = self.name()
		self._ptr += 1 + self._data[self._ptr] # description string
		self._ptr += self._ptr & 1 # align
		code = self.word() + self.word() # event class + event id
		# skip result
		self._ptr += 4 # datatype word
		self._ptr += 1 + self._data[self._ptr] # description string
		self._ptr += self._ptr & 1 # align
		self._ptr += 2 # flags integer
		# skip direct parameter
		self._ptr += 4 # datatype word
		self._ptr += 1 + self._data[self._ptr] # description string
		self._ptr += self._ptr & 1 # align
		self._ptr += 2 # flags integer
		#
		currentcommandargs = []
		# Note: overlapping command definitions (e.g. InDesign) should be processed as follows:
		# - If their names and codes are the same, only the last definition is used; other definitions are ignored and will not compile.
		# - If their names are the same but their codes are different, only the first definition is used; other definitions are ignored and will not compile.
		# - If a dictionary-defined command has the same name but different code to a built-in definition, escape its name so it doesn't conflict with the default built-in definition.
		if name not in self.commands or self.commands[name][1] == code:
			self.commands[name] =(name, code, currentcommandargs)
		# add labelled parameters
		for _ in range(self.integer()):
			name = self.name()
			self._ptr += self._ptr & 1 # align
			code = self.word()
			self._ptr += 4 # datatype word
			self._ptr += 1 + self._data[self._ptr] # description string
			self._ptr += self._ptr & 1 # align
			self._ptr += 2 # flags integer
			currentcommandargs.append((name, code))
	
	
	def parseclass(self):
		name = self.name()
		self._ptr += self._ptr & 1 # align
		code = self.word()
		self._ptr += 1 + self._data[self._ptr] # description string
		self._ptr += self._ptr & 1 # align
		isplural = False
		for _ in range(self.integer()): # properties
			propname = self.name()
			self._ptr += self._ptr & 1 # align
			propcode = self.word()
			self._ptr += 4 # datatype word
			self._ptr += 1 + self._data[self._ptr] # description string
			self._ptr += self._ptr & 1 # align
			flags = self.integer()
			if propcode != kae.pInherits: # not a superclass definition (see kAEInheritedProperties)
				if flags & 1: # indicates class name is plural (see kAESpecialClassProperties)
					isplural = True
				elif ((propname, propcode)) not in self._foundproperties:
					self.properties.append((propname, propcode)) # preserve ordering
					self._foundproperties.add((propname, propcode))
		for _ in range(self.integer()): # skip elements
			self._ptr += 4 # code word
			count = self.integer()
			self._ptr += 4 * count # reference forms
		if isplural:
			if ((name, code)) not in self._foundelements:
				self.elements.append((name, code))
				self._foundelements.add((name, code))
				self._foundelementcodes.add(code)
		else:
			if ((name, code)) not in self._foundclasses:
				self.classes.append((name, code))
				self._foundclasses.add((name, code))
				self._foundclasscodes.add(code)
		self._spareclassnames[code] = name
	
	
	def parsecomparison(self): # comparison info isn't used
		self._ptr += 1 + self._data[self._ptr] # name string
		self._ptr += self._ptr & 1 # align
		self._ptr += 4 # code word
		self._ptr += 1 + self._data[self._ptr] # description string
		self._ptr += self._ptr & 1 # align
	
	
	def parseenumeration(self): 
		self._ptr += 4 # code word
		for _ in range(self.integer()): # enumerators
			name = self.name()
			self._ptr += self._ptr & 1 # align
			code = self.word()
			self._ptr += 1 + self._data[self._ptr] # description string
			self._ptr += self._ptr & 1 # align
			if ((name, code)) not in self._foundenumerators:
				self.enumerators.append((name, code))
				self._foundenumerators.add((name, code))

	
	def parsesuite(self):
		self._ptr += 1 + self._data[self._ptr] # name string
		self._ptr += 1 + self._data[self._ptr] # description string
		self._ptr += self._ptr & 1 # align
		self._ptr += 4 # code word
		self._ptr += 4 # level, version integers
		for fn in [self.parsecommand, self.parseclass, self.parsecomparison, self.parseenumeration]:
			for _ in range(self.integer()):
				fn()
	
	#######
	# Public
	
	def parse(self, aetes):
		for aete in aetes:
			if isinstance(aete, ae.AEDesc) and aete.type in [kae.typeAETE, kae.typeAEUT] and aete.data:
				self._data = aete.data
				self._ptr = 6 # version, language, script integers
				for _ in range(self.integer()):
					self.parsesuite()
		# singular names are normally used in the classes table and plural names in the elements table. However, if an aete defines a singular name but not a plural name then the missing plural name is substituted with the singular name; and vice-versa if there's no singular equivalent for a plural name.
		missingelements = self._foundclasscodes - self._foundelementcodes
		missingclasses = self._foundelementcodes - self._foundclasscodes
		for code in missingelements:
			self.elements.append((self._spareclassnames[code], code))
		for code in missingclasses:
			self.classes.append((self._spareclassnames[code], code))
		return (self.classes, self.enumerators, self.properties, self.elements, list(self.commands.values()))


class LittleEndianParser(Parser):

	def word(self):
		"""Read a 4-byte string (really a long, but represented as an 4-character 8-bit string for readability)."""
		self._ptr += 4
		return self._data[self._ptr - 1:self._ptr - 5:-1] # little-endian


######################################################################
# PUBLIC
######################################################################

def buildtablesforaetes(aetes):
	"""
		aetes : list of aem.ae.AEDesc
	"""
	if pack("H", 1) == b'\x00\x01': # is it big-endian?
		return Parser().parse(aetes)
	else:
		return LittleEndianParser().parse(aetes)





"""py3-appscript -- High-level Mac OS X application scripting support for Python 3.1+. """

__version__ = '1.1.0'

__all__ = ['ApplicationNotFoundError', 'CommandError', 'CantLaunchApplicationError', 
		'app','con', 'its', 'k','mactypes']

from aem.findapp import ApplicationNotFoundError
from aem import CantLaunchApplicationError
from .reference import app, CommandError
from .genericreference import con, its
from .keywordwrapper import k
from aem import mactypes

# The following classes are exposed for occasional typechecking purposes. To avoid excess 
# namespace pollution they aren't added to the parent namespace when 'from appscript import *'
#  is used, so must be referred to like this [e.g.]:
#
# import appscript
# isinstance(obj, appscript.Reference)

from .reference import Command, Reference, Application, GenericApp
from .genericreference import GenericReference
from .keywordwrapper import Keyword


"""aemreference -- An object-oriented API for constructing object specifiers. """

import struct
from . import ae, kae

######################################################################
# SUPPORT FUNCTIONS
######################################################################

###################################
# packing support

if struct.pack("h", 1) == b'\x00\x01': # host is big-endian

	def packtype(code):
		return ae.newdesc(kae.typeType, code)
	
	def packabsoluteordinal(code): 
		return ae.newdesc(kae.typeAbsoluteOrdinal, code)
	
	def packenum(code):
		return ae.newdesc(kae.typeEnumeration, code)

else: # host is small-endian

	def packtype(code):
		return ae.newdesc(kae.typeType, code[::-1])
	
	def packabsoluteordinal(code): 
		return ae.newdesc(kae.typeAbsoluteOrdinal, code[::-1])
	
	def packenum(code):
		return ae.newdesc(kae.typeEnumeration, code[::-1])


def packlistas(type, lst):
	desc = ae.newrecord().coerce(type)
	for key, value in lst:
		desc.setparam(key, value)
	return desc


###################################
# comparison support

class _CollectComparable:
	def __init__(self):
		self.result = []
	
	def __getattr__(self, name):
		self.result.append(name)
		return self
	
	def __call__(self, *args):
		self.result.append(args)
		return self


######################################################################
# BASE CLASS
######################################################################

class Query(object):
	"""Base class for all specifier and testclause classes."""
	
	def __hash__(self):
		"""References are immutable, so may be used as dictionary keys."""
		return hash(repr(self))
	
	def __ne__(self, v):
		"""References may be compared for equality."""
		return not (self == v)
	
	def __eq__(self, v):
		"""References may be compared for equality."""
		return self is v or (
				self.__class__ == v.__class__ and 
				self.AEM_comparable() == v.AEM_comparable())
	
	def AEM_comparable(self):
		collector = _CollectComparable()
		self.AEM_resolve(collector)
		val = collector.result
		self.AEM_comparable = lambda: val
		return val


######################################################################
# BASE CLASS FOR ALL REFERENCE FORMS
######################################################################

class Specifier(Query):
	"""Base class for all object specifier classes."""
	
	def __init__(self, container, key):
		self._container = container
		self._key = key
		
	def AEM_root(self):
		# Get reference's root node. Used by range and filter specifiers when determining type of reference
		# passed as argument(s): range specifiers take absolute (app-based) and container (con-based)
		# references; filter specifiers require an item (its-based) reference.
		return self._container.AEM_root()
	
	def AEM_trueself(self):
		# Called by specifier classes when creating a reference to sub-element(s) of the current reference.
		# - An AllElements specifier (which contains 'want', 'form', 'seld' and 'from' values) will return an UnkeyedElements object (which contains 'want' and 'from' data only). The new specifier object  (ElementByIndex, ElementsByRange, etc.) wraps itself around this stub and supply its own choice of 'form' and 'seld' values.
		# - All other specifiers simply return themselves. 
		#
		#This sleight-of-hand allows foo.elements('bar ') to produce a legal reference to all elements, so users don't need to write foo.elements('bar ').all to achieve the same goal. This isn't a huge deal for aem, but makes a significant difference to the usability of user-friendly wrappers like appscript.
		return self
	
	def AEM_packself(self, codecs):
		# Pack this Specifier; called by codecs.
		desc = self._packself(codecs)
		self.AEM_packself = lambda codecs: desc # once packed, reuse this AEDesc for efficiency
		return desc


######################################################################
# INSERTION POINT REFERENCE FORM
######################################################################

class InsertionSpecifier(Specifier):
	"""Form: allelementsref.beginning/end, elementsref.before/after
		A reference to an element insertion point.
	"""
	def __init__(self, container, key, keyname):
		Specifier.__init__(self, container, key)
		self._keyname = keyname
	
	def __repr__(self):
		return '{!r}.{}'.format(self._container, self._keyname)
	
	def _packself(self, codecs):
		return packlistas(kae.typeInsertionLoc, [
				(kae.keyAEObject, self._container.AEM_packself(codecs)), 
				(kae.keyAEPosition, self._key),
				])
	
	def AEM_resolve(self, obj):
		return getattr(self._container.AEM_resolve(obj), self._keyname)


######################################################################
# BASE CLASS FOR ALL OBJECT REFERENCE FORMS
######################################################################

class _PositionSpecifier(Specifier):
	"""All property and element reference forms inherit from this class.
	
	Note that comparison and logic 'operator' methods are implemented on this class - these are only for use in constructing its-based references and shouldn't be used on app- and con-based references. Aem doesn't enforce this rule itself so as to minimise runtime overhead (the target application will raise an error if the user does something foolish).
	"""
	
	_kBeginning = packenum(kae.kAEBeginning)
	_kEnd = packenum(kae.kAEEnd)
	_kBefore = packenum(kae.kAEBefore)
	_kAfter = packenum(kae.kAEAfter)
	_kPrevious = packenum(kae.kAEPrevious)
	_kNext = packenum(kae.kAENext)
	
	def __init__(self, wantcode, container, key):
		self.AEM_want = wantcode
		Specifier.__init__(self, container, key)
	
	def __repr__(self):
		return '{!r}.{}({!r})'.format(self._container, self._kBy, self._key)
	
	def _packself(self, codecs):
		return packlistas(kae.typeObjectSpecifier, [
				(kae.keyAEDesiredClass, packtype(self.AEM_want)),
				(kae.keyAEKeyForm, self._kKeyForm),
				(kae.keyAEKeyData, self._packkey(codecs)),
				(kae.keyAEContainer, self._container.AEM_packself(codecs)),
				])
	
	# Comparison tests; these should only be used on its-based references:
	
	def gt(self, val):
		"""gt(anything) --> is greater than test"""
		return GreaterThan(self, val)
		
	def ge(self, val):
		"""ge(anything) --> is greater than or equals test"""
		return GreaterOrEquals(self, val)
	
	def eq(self, val):
		"""eq(anything) --> equals test"""
		return Equals(self, val)
	
	def ne(self, val):
		"""ne(anything) --> does not equal test"""
		return NotEquals(self, val)
	
	def lt(self, val):
		"""lt(anything) --> is less than test"""
		return LessThan(self, val)
	
	def le(self, val):
		"""le(anything) --> is less than or equals test"""
		return LessOrEquals(self, val)
	
	def beginswith(self, val):
		"""beginswith(anything) --> begins with test"""
		return BeginsWith(self, val)
	
	def endswith(self, val):
		"""endswith(anything) --> ends with test"""
		return EndsWith(self, val)
	
	def contains(self, val):
		"""contains(anything) --> contains test"""
		return Contains(self, val)
	
	def isin(self, val):
		"""isin(anything) --> isin test"""
		return IsIn(self, val)
	
	# Insertion references can be used on any kind of element reference, and also on property references where the property represents a one-to-one relationship, e.g. textedit.documents[1].text.end is valid:
		
	beginning = property(lambda self: InsertionSpecifier(self, self._kBeginning, 'beginning'), doc="beginning --> insertion location")
	end = property(lambda self: InsertionSpecifier(self, self._kEnd, 'end'), doc="end --> insertion location")
	before = property(lambda self: InsertionSpecifier(self, self._kBefore, 'before'), doc="before --> insertion location")
	after = property(lambda self: InsertionSpecifier(self, self._kAfter, 'after'), doc="after --> insertion location")
	
	# Property and element references can be used on any type of object reference:
	
	def property(self, propertycode):
		"""property(propertycode) --> property"""
		return Property(kae.cProperty, self, propertycode)
	
	def userproperty(self, name):
		"""property(name) --> property"""
		return UserProperty(kae.cProperty, self, name)
	
	def elements(self, elementcode):
		"""elements(elementcode) --> all elements"""
		return AllElements(elementcode, self)
	
	# Relative position references are unlikely to work on one-to-one relationships - but what the hey, it simplifies the class structure a bit.
	
	def previous(self, elementcode):
		"""previous(elementcode) --> element"""
		return ElementByRelativePosition(elementcode, self, self._kPrevious, 'previous')
	
	def next(self, elementcode):
		"""next(elementcode) --> element"""
		return ElementByRelativePosition(elementcode, self, self._kNext, 'next')


######################################################################
# PROPERTY REFERENCE FORMS
######################################################################

class Property(_PositionSpecifier):
	"""Form: ref.property(code)
		A reference to an application-defined property, where code is the code identifying the property.
	"""
	_kBy = 'property'
	_kKeyForm = packenum(kae.formPropertyID)
	
	def _packkey(self, codecs):
		return packtype(self._key)
	
	def AEM_resolve(self, obj):
		return self._container.AEM_resolve(obj).property(self._key)


class UserProperty(_PositionSpecifier):
	"""Form: ref.userproperty(name)
		A reference to a user-defined property, where name is a string representing the property's name. 
		
		Scriptable applications shouldn't use this reference form, but OSA script applets can.
		Note that OSA languages may have additional rules regarding case sensitivity/conversion.
	"""
	_kBy = 'userproperty'
	_kKeyForm = packenum(kae.formUserPropertyID)
	
	def _packkey(self, codecs):
		return codecs.pack(self._key).coerce(kae.typeChar)
	
	def AEM_resolve(self, obj):
		return self._container.AEM_resolve(obj).userproperty(self._key)


######################################################################
# ELEMENT REFERENCE FORMS
######################################################################

###################################
# Single elements

class _SingleElement(_PositionSpecifier):
	"""Base class for all single element specifiers."""
	
	def __init__(self, wantcode, container, key):
		# Notes: when byindex, byname, byid, first, middle, last or any is called on an AllElements object, we want to 'strip' the AllElements object away and use the underlying UnkeyedElements object as our 'container' instead. AEM_trueself returns the UnkeyedElements object when called on an AllElements object; in all other cases it returns the same object it was called on.
		_PositionSpecifier.__init__(self, wantcode, container.AEM_trueself(), key)
	
	def _packkey(self, codecs):
		return codecs.pack(self._key)
	
	def AEM_resolve(self, obj):
		return getattr(self._container.AEM_resolve(obj), self._kBy)(self._key)


#######

class ElementByName(_SingleElement):
	"""Form: elementsref.byname(text)
		A reference to a single element by its name, where text is string or unicode.
	"""
	_kBy = 'byname'
	_kKeyForm = packenum(kae.formName)


class ElementByIndex(_SingleElement):
	"""Form: elementsref.byindex(i)
		A reference to a single element by its index, where i is a non-zero whole number.
	"""
	_kBy = 'byindex'
	_kKeyForm = packenum(kae.formAbsolutePosition)


class ElementByID(_SingleElement):
	"""Form: elementsref.byid(anything)
		A reference to a single element by its id.
	"""
	_kBy = 'byid'
	_kKeyForm = packenum(kae.formUniqueID)

##

class ElementByOrdinal(_SingleElement):
	"""Form: elementsref.first/middle/last/any
		A reference to first/middle/last/any element.
	"""
	_kKeyForm = packenum(kae.formAbsolutePosition)
	
	def __init__(self, wantcode, container, key, keyname):
		self._keyname = keyname
		_SingleElement.__init__(self, wantcode, container, key)
	
	def __repr__(self):
		return '{!r}.{}'.format(self._container, self._keyname)
	
	def AEM_resolve(self, obj):
		return getattr(self._container.AEM_resolve(obj), self._keyname)


class ElementByRelativePosition(_SingleElement):
	"""Form: elementsref.previous/next(code)
		A relative reference to previous/next element, where code
		is the class code of element to get.
	"""
	_kKeyForm = packenum(kae.formRelativePosition)
	
	def __init__(self, wantcode, container, key, keyname):
		# Note: this method overrides _SingleElement.__init__() since we want to keep any AllElements container references as-is, not sub-select them.
		self._keyname = keyname
		_PositionSpecifier.__init__(self, wantcode, container, key)
	
	def __repr__(self):
		return '{!r}.{}({!r})'.format(self._container, self._keyname, self.AEM_want)
	
	def AEM_resolve(self, obj):
		return getattr(self._container.AEM_resolve(obj), self._keyname)(self.AEM_want)


###################################
# Multiple elements

class _MultipleElements(_PositionSpecifier):
	"""Base class for all multiple element specifiers."""
	
	_kFirst = packabsoluteordinal(kae.kAEFirst)
	_kMiddle = packabsoluteordinal(kae.kAEMiddle)
	_kLast = packabsoluteordinal(kae.kAELast)
	_kAny = packabsoluteordinal(kae.kAEAny)
	
	first = property(lambda self: ElementByOrdinal(self.AEM_want, self, self._kFirst, 'first'), doc="first --> element")
	middle = property(lambda self: ElementByOrdinal(self.AEM_want, self, self._kMiddle, 'middle'), doc="middle --> element")
	last = property(lambda self: ElementByOrdinal(self.AEM_want, self, self._kLast, 'last'), doc="last --> element")
	any = property(lambda self: ElementByOrdinal(self.AEM_want, self, self._kAny, 'any'), doc="any --> element")
	
	def byname(self, name):
		"""byname(name) --> element"""
		return ElementByName(self.AEM_want, self, name)
	
	def byindex(self, index):
		"""byindex(index) --> element"""
		return ElementByIndex(self.AEM_want, self, index)
	
	def byid(self, id):
		"""byid(id) --> element"""
		return ElementByID(self.AEM_want, self, id)
	
	def byrange(self, start, stop):
		"""byrange(start, stop) --> elements"""
		return ElementsByRange(self.AEM_want, self, (start, stop))
	
	def byfilter(self, expression):
		"""byfilter(expression) --> elements"""
		return ElementsByFilter(self.AEM_want, self, expression)


#######

class ElementsByRange(_MultipleElements):
	"""Form: elementsref.range(start, stop)
		A reference to a range of elements, where start and stop are relative references 
		to the first and last elements in range (see also 'con').
	"""
	_kKeyForm = packenum(kae.formRange)
	
	def __init__(self, wantcode, container, key):
		_MultipleElements.__init__(self, wantcode, container.AEM_trueself(), key)
	
	def __repr__(self):
		return '{!r}.byrange({!r}, {!r})'.format(self._container, *self._key)

	def _packkey(self, codecs):
		rangeselectors = []
		for key, selector in [(kae.keyAERangeStart, self._key[0]), (kae.keyAERangeStop, self._key[1])]:
			if isinstance(selector, Specifier):
				rangeselectors.append([key, codecs.pack(selector)])
			elif isinstance(selector, str):
				rangeselectors.append([key, codecs.pack(con.elements(self.AEM_want).byname(selector))])
			else:
				rangeselectors.append([key, codecs.pack(con.elements(self.AEM_want).byindex(selector))])
		return packlistas(kae.typeRangeDescriptor, rangeselectors)
	
	def AEM_resolve(self, obj):
		return self._container.AEM_resolve(obj).byrange(*self._key)


class ElementsByFilter(_MultipleElements):
	"""Form: elementsref.filter(expr)
		A reference to all elements that match a condition, where expr 
		is a relative reference to the object being tested (see also 'its').
	"""
	_kKeyForm = packenum(kae.formTest)
	
	def __init__(self, wantcode, container, key):
		if not isinstance(key, Test):
			raise TypeError('Not a test specifier: {!r}'.format(key))
		_MultipleElements.__init__(self, wantcode, container.AEM_trueself(), key)
	
	def __repr__(self):
		return '{!r}.byfilter({!r})'.format(self._container, self._key)

	def _packkey(self, codecs):
		return codecs.pack(self._key)
	
	def AEM_resolve(self, obj):
		return self._container.AEM_resolve(obj).byfilter(self._key)


class AllElements(_MultipleElements):
	"""Form: ref.elements(code)
		A reference to all elements of container, where code is elements' class code.
	"""
	_kKeyForm = packenum(kae.formAbsolutePosition)
	_kAll = packabsoluteordinal(kae.kAEAll)
	
	def __init__(self, wantcode, container):
		# An AllElements object is a wrapper around an UnkeyedElements object; when selecting one or more of these elements, the AllElements wrapper is skipped and the UnkeyedElements object is used as the 'container' for the new specifier.
		_PositionSpecifier.__init__(self, wantcode, UnkeyedElements(wantcode, container), self._kAll)
	
	def __repr__(self):
		return repr(self._container)
	
	def _packkey(self, codecs):
		return self._kAll
	
	def AEM_trueself(self): # override default implementation to return the UnkeyedElements object stored inside of this AllElements instance
		return self._container
	
	def AEM_resolve(self, obj):
		return self._container.AEM_resolve(obj) # forward to UnkeyedElements


######################################################################
# SHIMS
######################################################################

###################################
# Multiple element shim

class UnkeyedElements(Specifier):
	"""
		A partial elements reference, containing element code but no keyform/keydata. A shim.
		User is never exposed to this class directly. 
		
		The goal here is simple: to allow users to write 'x.elements(code)' to refer to all elements, 
		instead of the clumsier 'x.elements(code).all', as well as stuff like x.elements.first,
		x.elements.byindex(i), x.elements(code).elements(code), x.elements(code).byfilter(f).first, 
		and so on.
		
		Here's how it behaves:
		
		- Calling a reference's element() method initially returns an UnkeyedElements instance
		wrapped inside an AllElements instance, e.g. app.elements('docu'). 
		
		- Calling an element selection method on the AllElements instance, 
		e.g. app.elements('docu').byindex(1), strips away the AllElements instance to obtain the
		UnkeyedElements instance which is then used as the foundation for this new specifier.
		
		(There is one exception: ElementByRelativePosition. This keeps the AllElements reference
		intact, since it identifies a sibling of the currently specified elements.)
		
		- Calling property() and element() methods on any reference does no stripping, nor does calling
		element selection methods on other multi-item specifiers (ElementsByRange, ElementsByFilter).
		In both cases, we're stepping down a level in the object model so want the AllElements reference
		to objects at this level intact.
		
		This extra work also makes the higher-level appscript wrapper simpler to implement, since
		the behaviour here is the same as there. While one could implement separate allelements,
		firstelement, elementbyindex, elementbyrange, etc. methods in the aem layer and then do the 
		shimming in the appscript layer, this way is more consistent.
	"""
	
	def __init__(self, wantcode, container):
		self.AEM_want = wantcode
		self._container = container
	
	def __repr__(self):
		return '{!r}.elements({!r})'.format(self._container, self.AEM_want)
	
	def AEM_packself(self, codecs):
		return self._container.AEM_packself(codecs) # forward to container specifier
	
	def AEM_resolve(self, obj):
		return self._container.AEM_resolve(obj).elements(self.AEM_want)


###################################
# Unresolved reference

class DeferredSpecifier(Query):	
	"""Deferred specifier; used to represent unresolved container references that may need to be resolved later. A performance optimisation.
	
	When unpacking specifier AEDescs of typeObjectSpecifier, if the topmost AEDesc is of formPropertyID, formAbsolutePosition, formName and formUniqueID (the simplest and most commonly used forms), its container AEDesc isn't unpacked immediately; instead, it's placed in a DeferredSpecifier instance and is only unpacked if actually needed (e.g. when __repr__ or AEM_resolve is called). 
	
	This makes the implementation a little more complex, but gives an approximately 2x speed up when unpacking references. Repacking these references is also faster as the original AEDesc is retained instead of being repacked from scratch.
	"""

	def __init__(self, desc, codecs):
		self._desc = desc
		self._codecs = codecs
	
	def _realref(self):
		ref = self._codecs.unpack(self._desc) or self._codecs.app
		if not isinstance(ref, Query):
			if ref is None:
				ref = self._codecs.app
			else:
				ref = customroot(ref)
		self._realref = lambda:ref
		return ref
	
	def AEM_trueself(self):
		return self
		
	def __repr__(self):
		return repr(self._realref())
	
	def __eq__(self, v):
		return self._realref() == v
	
	def __hash__(self):
		return hash(self._realref())
	
	def AEM_root(self):
		return self._realref().AEM_root()
	
	def AEM_resolve(self, obj):
		return self._realref().AEM_resolve(obj)
		

######################################################################
# TEST CLAUSES
######################################################################

###################################
# Base class

class Test(Query):
	"""Base class for all comparison and logic test classes (Equals, NotEquals, AND, OR, etc.)."""

	# Logical tests.
	def AND(self, operand2, *operands):
		"""AND(test,...) --> logical AND test"""
		return AND((self, operand2) + operands)
		
	def OR(self, operand2, * operands):
		"""OR(test,...) --> logical OR test"""
		return OR((self, operand2) + operands)
	
	NOT = property(lambda self: NOT((self,)), doc="NOT --> logical NOT test")


###################################
# Comparison tests

class _ComparisonTest(Test):
	"""Subclassed by comparison test classes."""
	def __init__(self, operand1, operand2):
		self._operand1 = operand1
		self._operand2 = operand2
	
	def __repr__(self):
		return '{!r}.{}({!r})'.format(self._operand1, self._name, self._operand2)

	def AEM_resolve(self, obj):
		return getattr(self._operand1.AEM_resolve(obj), self._name)(self._operand2)

	def AEM_packself(self, codecs):
		return packlistas(kae.typeCompDescriptor, [
				(kae.keyAEObject1, codecs.pack(self._operand1)), 
				(kae.keyAECompOperator, self._operator), 
				(kae.keyAEObject2, codecs.pack(self._operand2))
				])

##

class GreaterThan(_ComparisonTest):
	_name = 'gt'
	_operator = packenum(kae.kAEGreaterThan)

class GreaterOrEquals(_ComparisonTest):
	_name = 'ge'
	_operator = packenum(kae.kAEGreaterThanEquals)

class Equals(_ComparisonTest):
	_name = 'eq'
	_operator = packenum(kae.kAEEquals)

class NotEquals(Equals):
	_name = 'ne'
	_operatorNOT = packenum(kae.kAENOT)
	
	def AEM_packself(self, codecs):
		return self._operand1.eq(self._operand2).NOT.AEM_packself(codecs)

class LessThan(_ComparisonTest):
	_name = 'lt'
	_operator = packenum(kae.kAELessThan)

class LessOrEquals(_ComparisonTest):
	_name = 'le'
	_operator = packenum(kae.kAELessThanEquals)

class BeginsWith(_ComparisonTest):
	_name = 'beginswith'
	_operator = packenum(kae.kAEBeginsWith)

class EndsWith(_ComparisonTest):
	_name = 'endswith'
	_operator = packenum(kae.kAEEndsWith)

class Contains(_ComparisonTest):
	_name = 'contains'
	_operator = packenum(kae.kAEContains)

class IsIn(Contains):
	_name = 'isin'

	def AEM_packself(self, codecs):
		return packlistas(kae.typeCompDescriptor, [
				(kae.keyAEObject1, codecs.pack(self._operand2)), 
				(kae.keyAECompOperator, self._operator), 
				(kae.keyAEObject2, codecs.pack(self._operand1))
				])


###################################
# Logical tests

class _LogicalTest(Test):
	"""Subclassed by logical test classes."""
	def __init__(self, operands):
		self._operands = operands
		
	def __repr__(self):
		return '{!r}.{}({})'.format(self._operands[0], self._name, repr(list(self._operands[1:]))[1:-1])
	
	def AEM_resolve(self, obj):
		return getattr(self._operands[0].AEM_resolve(obj), self._name)(*self._operands[1:])
	
	def AEM_packself(self, codecs):
		return packlistas(kae.typeLogicalDescriptor, [
				(kae.keyAELogicalOperator, self._operator), 
				(kae.keyAELogicalTerms, codecs.pack(self._operands)),
				])

##

class AND(_LogicalTest):
	_operator = packenum(kae.kAEAND)
	_name = 'AND'


class OR(_LogicalTest):
	_operator = packenum(kae.kAEOR)
	_name = 'OR'


class NOT(_LogicalTest):
	_operator = packenum(kae.kAENOT)
	_name = 'NOT'
		
	def __repr__(self):
		return '{!r}.NOT'.format(self._operands[0])
	
	def AEM_resolve(self, obj):
		return self._operands[0].AEM_resolve(obj).NOT


######################################################################
# REFERENCE ROOTS
######################################################################

###################################
# Base class


class ReferenceRoot(_PositionSpecifier):
	def __init__(self):
		pass
	
	def __repr__(self):
		return self._kName
	
	def _packself(self, codecs):
		return self._kType
	
	def AEM_root(self):
		return self
	
	def AEM_resolve(self, obj):
		return getattr(obj, self._kName)


###################################
# Concrete classes

class ApplicationRoot(ReferenceRoot):
	"""Form: app
		Reference base; represents an application's application object. Used to construct full references.
	"""
	_kName = 'app'
	_kType = ae.newdesc(kae.typeNull, b'')


class CurrentContainer(ReferenceRoot):
	"""Form: con
		Reference base; represents elements' container object. Used to construct by-range references.
	"""
	_kName = 'con'
	_kType = ae.newdesc(kae.typeCurrentContainer, b'')


class ObjectBeingExamined(ReferenceRoot):
	"""Form: its
		Reference base; represents an element to be tested. Used to construct by-filter references.
	"""
	_kName = 'its'
	_kType = ae.newdesc(kae.typeObjectBeingExamined, b'')


class CustomRoot(ReferenceRoot):
	"""Form: customroot(obj)
		Reference base; represents an arbitrary root object, e.g. an AEAddressDesc in a fully qualified reference.
	"""

	def __init__(self, rootObj):
		ReferenceRoot.__init__(self)
		self._rootObj = rootObj
	
	def __repr__(self):
		return 'customroot({!r})'.format(self._rootObj)
	
	def _packself(self, codecs):
		return codecs.pack(self._rootObj)
	
	def AEM_resolve(self, obj):
		return obj.customroot(self._rootObj)


###################################
# Reference root objects; use these constants to construct new specifiers, e.g. app.property('pnam')

app = ApplicationRoot()
con = CurrentContainer()
its = ObjectBeingExamined()
customroot = CustomRoot




"""osax.py -- Allows scripting additions (a.k.a. OSAXen) to be called from Python. """

from xml.etree import ElementTree
import io, string

from appscript import *
from appscript import reference, terminology
from appscript.reservedkeywords import kReservedKeywords
import aem


__all__ = ['OSAX', 'scriptingadditions', 'ApplicationNotFoundError', 'CommandError', 'k', 'mactypes']


######################################################################
# PRIVATE
######################################################################


class TreeBuilderIgnoreDoctype(ElementTree.TreeBuilder):
	""" Suppress XMLParser.doctype() deprecation warnings in ElementTree 1.3 """
	
	def doctype(self, name, pubid, system):
		pass


class SdefParser:
	""" Parse scripting addition sdef.
		
		Note:
		
		- OSAX terminology can only be retrieved via OSAGetAppTerminology or OSACopyScriptingDefinition;
			while the former is more reliable, it is deprecated in 32-bit and absent in 64-bit Carbon APIs.
		
		- Four char codes that contain non-printing codes in aetes (e.g. the 'caution' enum in StandardAdditions has code '\x00\x00\x00\x02') do not convert to valid code attributes in sdefs.
		
		- xi:include or class-extension elements are not supported, but osaxen are unlikely to use these anyway.
		
		- Synonyms are not supported. (While possible, it would require additional work to ensure that synonym names/codes don't accidentally mask the primary terms.)
	"""
	
	_keywordcache = {}
	_reservedwords = set(kReservedKeywords)
	_specialconversions = {
			' ': '_',
			'-': '_',
			'&': 'and',
			'/': '_',
	}
	_legalchars = string.ascii_letters + '_'
	_alphanum = _legalchars + string.digits
	
	commands = property(lambda self: list(self._commands.values()))

	def __init__(self):
		self.classes, self.enums, self.properties, self.elements, self._commands = [], [], [], [], {}
	
	def _name(self, s):
		if s not in self._keywordcache:
			legal = self._legalchars
			res = ''
			for c in s:
				if c in legal:
					res += c
				elif c in self._specialconversions:
					res += self._specialconversions[c]
				else:
					if res == '':
						res = '_' # avoid creating an invalid identifier
					res += '0x{:X}'.format(ord(c))
				legal = self._alphanum
			if res in self._reservedwords or res.startswith('_') or res.startswith('AS_'):
				res += '_'
			self._keywordcache[s] = str(res)
		return self._keywordcache[s]
	
	def _code(self, s):
		return s.encode('macroman')
	
	def _addnamecode(self, node, collection):
		name = self._name(node.get('name'))
		code = self._code(node.get('code'))
		if name and len(code) == 4 and (name, code) not in collection:
			collection.append((name, code))
	
	def _addcommand(self, node):
		name = self._name(node.get('name'))
		code = self._code(node.get('code'))
		parameters = []
		# Note: overlapping command definitions (e.g. 'path to') should be processed as follows:
		# - If their names and codes are the same, only the last definition is used; other definitions are ignored and will not compile.
		# - If their names are the same but their codes are different, only the first definition is used; other definitions are ignored and will not compile.
		if name and len(code) == 8 and (name not in self._commands or self._commands[name][1] == code):
			self._commands[name] =(name, code, parameters)
			for pnode in node.findall('parameter'):
				self._addnamecode(pnode, parameters)
	
	def parse(self, xml):
		""" Extract name-code mappings from an sdef.
		
			xml : bytes | str -- sdef data
		"""
		xml = ElementTree.parse(io.BytesIO(xml), ElementTree.XMLParser(target=TreeBuilderIgnoreDoctype()))
		for suite in xml.findall('suite'):
			for node in suite:
				try:
					if node.tag in ['command', 'event']:
						self._addcommand(node)
					elif node.tag in ['class', 'record-type', 'value-type']:
						self._addnamecode(node, self.classes)
						for prop in node.findall('property'):
							self._addnamecode(prop, self.properties)
						if node.tag == 'class': # elements
							name = self._name(node.get('plural', 
									node.get('name', '') + 's'))
							code = self._code(node.get('code'))
							if name and name != 's' and len(code) == 4:
								self.elements.append((name, code))
					elif node.tag == 'enumeration':
						for enum in node.findall('enumerator'):
							self._addnamecode(enum, self.enums)
				except:
					pass # ignore problem definitions
	
	def parsefile(self, path):
		""" Extract name-code mappings from an sdef.
		
			path :  str -- path to .sdef file
		"""
		self.parse(aem.ae.copyscriptingdefinition(path))


##


kStandardAdditionsEnums = [ # codes from StandardAdditions aete
		('stop', b'\x00\x00\x00\x00'),
		('note', b'\x00\x00\x00\x01'),
		('caution', b'\x00\x00\x00\x02')]


######################################################################


_osaxcache = {} # a dict of form: {'osax name': ['/path/to/osax', cached_terms_or_None], ...}

_osaxnames = [] # names of all currently available osaxen

def _initcaches():
		_se = aem.Application(aem.findapp.byid('com.apple.systemevents'))
		for domaincode in [b'flds', b'fldl', b'fldu']:
			osaxen = aem.app.property(domaincode).property(b'$scr').elements(b'file').byfilter(
					aem.its.property(b'asty').eq('osax').OR(aem.its.property(b'extn').eq('osax')))
			if _se.event(b'coredoex', {b'----': osaxen.property(b'pnam')}).send(): # domain has ScriptingAdditions folder
				names = _se.event(b'coregetd', {b'----': osaxen.property(b'pnam')}).send()
				paths = _se.event(b'coregetd', {b'----': osaxen.property(b'posx')}).send()
				for name, path in zip(names, paths):
					if name.lower().endswith('.osax'): # remove name extension, if any
						name = name[:-5]
					if name.lower() not in _osaxcache:
						_osaxnames.append(name)
						_osaxcache[name.lower()] = [path, None]
		_osaxnames.sort()


######################################################################
# PUBLIC
######################################################################

def scriptingadditions():
	if not _osaxnames:
		_initcaches()
	return _osaxnames[:]


class OSAX(reference.Application):

	def __init__(self, osaxname='StandardAdditions', name=None, id=None, creator=None, pid=None, url=None, aemapp=None, terms=True):
		if not _osaxcache:
			_initcaches()
		self._osaxname = osaxname
		osaxname = osaxname.lower()
		if osaxname.endswith('.osax'):
			osaxname = osaxname[:-5]
		if terms == True:
			try:
				osaxpath, terms = _osaxcache[osaxname]
			except KeyError as e:
				raise ValueError("Scripting addition not found: {!r}".format(self._osaxname)) from e
			if not terms:
				p = SdefParser()
				p.parsefile(osaxpath)
				if osaxname == 'standardadditions':
					p.enums += kStandardAdditionsEnums # patch in missing codes for compatibility
				terms = _osaxcache[osaxname][1] = terminology.tablesformodule(p)
		reference.Application.__init__(self, name, id, creator, pid, url, aemapp, terms)
		try:
			self.AS_appdata.target().event(b'ascrgdut').send(300) # make sure target application has loaded event handlers for all installed OSAXen
		except aem.EventError as e:
			if e.errornumber != -1708: # ignore 'event not handled' error
				raise
		def _help(*args):
			raise NotImplementedError("Built-in help isn't available for scripting additions.")
		self.AS_appdata.help = _help
		
	def __str__(self):
		if self.AS_appdata.constructor == 'current':
			return 'OSAX({!r})'.format(self._osaxname)
		elif self.AS_appdata.constructor == 'path':
			return 'OSAX({!r}, {!r})'.format(self._osaxname, self.AS_appdata.identifier)
		else:
			return 'OSAX({!r}, {}={!r})'.format(self._osaxname, self.AS_appdata.constructor, self.AS_appdata.identifier)
			
	def __getattr__(self, name):
		command = reference.Application.__getattr__(self, name)
		if isinstance(command, reference.Command):
			def osaxcommand(*args, **kargs):
				try:
					return command(*args, **kargs)
				except CommandError as e:
					if int(e) == -1713: # 'No user interaction allowed' error (e.g. user tried to send a 'display dialog' command to a non-GUI python process), so convert the target process to a full GUI process and try again
						aem.ae.transformprocesstoforegroundapplication()
						self.activate()
						return command(*args, **kargs)
					raise
			return osaxcommand
		else:
			return command
		
	__repr__ = __str__


##

def dump(osaxname, modulepath):
	"""Dump scripting addition terminology data to Python module.
		osaxname : str -- name of installed scripting addition
		modulepath : str -- path to generated module
		
	Generates a Python module containing a scripting addition's basic terminology 
	(names and codes) as used by osax.
	
	Call the dump() function to dump faulty sdefs to Python module, e.g.:
	
		dump('MyOSAX', '/path/to/site-packages/myosaxglue.py')
	
	Patch any errors by hand, then import the patched module into your script 
	and pass it to osax's OSAX() constructor via its 'terms' argument, e.g.:
	
		from osax import *
		import myosaxglue
		
		myapp = OSAX('MyOSAX', terms=myosaxglue)
	"""
	if not _osaxnames:
		_initcaches()
	originalname = osaxname
	osaxname = osaxname.lower()
	if osaxname.endswith('.osax'):
		osaxname = osaxname[:-5]
	try:
		osaxpath, terms = _osaxcache[osaxname]
	except KeyError as e:
		raise ValueError("Scripting addition not found: {!r}".format(originalname)) from e
	p = SdefParser()
	p.parsefile(osaxpath)
	terminology.dumptables((p.classes, p.enums, p.properties, p.elements, p.commands), osaxpath, modulepath)




"""codecs -- Convert from Python to Apple Event Manager types and vice-versa. """

import struct, datetime, time
from codecs import BOM_UTF16_LE, BOM_UTF16_BE

from .ae import AEDesc, newdesc, newlist, newrecord
from .typewrappers import AEType, AEEnum, AEProp, AEKey
from . import aemreference, mactypes, kae


######################################################################
# PRIVATE
######################################################################

if struct.pack("h", 1) == b'\x00\x01': # host is big-endian
	fourcharcode = lambda code: code
	nativeutf16encoding = 'UTF-16BE'
else: # host is small-endian
	fourcharcode = lambda code: code[::-1]
	nativeutf16encoding = 'UTF-16LE'


class _Ordinal:
	def __init__(self, code):
		self.code = code

class _Range:
	def __init__(self, range):
		self.range = range


######################################################################
# PUBLIC
######################################################################


class UnitTypeCodecs:
	"""Provides pack and unpack methods for converting between mactypes.Units instances and AE unit types. Each Codecs instance is allocated its own UnitTypeCodecs instance.
	"""
	
	_defaultunittypes = [
		('centimeters', b'cmtr'),
		('meters', b'metr'),
		('kilometers', b'kmtr'),
		('inches', b'inch'),
		('feet', b'feet'),
		('yards', b'yard'),
		('miles', b'mile'),
		
		('square_meters', b'sqrm'),
		('square_kilometers', b'sqkm'),
		('square_feet', b'sqft'),
		('square_yards', b'sqyd'),
		('square_miles', b'sqmi'),
		
		('cubic_centimeters', b'ccmt'),
		('cubic_meters', b'cmet'),
		('cubic_inches', b'cuin'),
		('cubic_feet', b'cfet'),
		('cubic_yards', b'cyrd'),
		
		('liters', b'litr'),
		('quarts', b'qrts'),
		('gallons', b'galn'),
		
		('grams', b'gram'),
		('kilograms', b'kgrm'),
		('ounces', b'ozs '),
		('pounds', b'lbs '),
		
		('degrees_Celsius', b'degc'),
		('degrees_Fahrenheit', b'degf'),
		('degrees_Kelvin', b'degk'),
	]
	
	##
	
	def _defaultpacker(self, units, code): 
		return newdesc(code, struct.pack('d', units.value))
	
	def _defaultunpacker(self, desc, name):
		return mactypes.Units(struct.unpack('d', desc.data)[0], name)
	
	##
	
	def __init__(self):
		self._typebyname = {}
		self._typebycode = {}
		self.addtypes(self._defaultunittypes)
	
	def addtypes(self, typedefs):
		""" Add application-specific unit type definitions to this UnitTypeCodecs instance.
		
			typedefs is a list of tuples, where each tuple is of form:
				(typename, typecode, packer, unpacker)
			or:
				(typename, typecode)
			
			If optional packer and unpacker functions are omitted, default pack/unpack functions
			are used instead; these pack/unpack AEDesc data as a double precision float.
		"""
		for item in typedefs:
			if len(item) == 2:
				item = item + (self._defaultpacker, self._defaultunpacker)
			name, code, packer, unpacker = item
			self._typebyname[name] = (code, packer)
			self._typebycode[code] = (name, unpacker)
	
	def pack(self, val):
		if isinstance(val, mactypes.Units):
			try:
				code, packer = self._typebyname[val.type]
			except KeyError as e:
				raise TypeError('Unknown unit type: {!r}'.format(val)) from e
			else:
				return True, packer(val, code)
		else:
			return False, val
	
	def unpack(self, desc):
		if desc.type in self._typebycode:
			name, unpacker = self._typebycode[desc.type]
			return True, unpacker(desc, name)
		else:
			return False, desc



######################################################################


class Codecs:
	"""Convert between Python and Apple event data types.
	Clients may add additional encoders/decoders and/or subclass to suit their needs.
	"""
	
	# Constants
	
	kNullDesc = newdesc(kae.typeNull, '')
	kMacEpoch = datetime.datetime(1904, 1, 1) # used in packing datetime objects as AEDesc typeLongDateTime
	kMacEpochT = time.mktime(kMacEpoch.timetuple())
	kShortMacEpoch = kMacEpoch.date() # used in packing date objects as AEDesc typeLongDateTime

	kTrueDesc = newdesc(kae.typeTrue, '')
	kFalseDesc = newdesc(kae.typeFalse, '')
	
	#######
	# tables to map AE codes to aem method names
	
	kInsertionLocSelectors = {
			fourcharcode(kae.kAEBefore): 'before', 
			fourcharcode(kae.kAEAfter): 'after', 
			fourcharcode(kae.kAEBeginning): 'beginning', 
			fourcharcode(kae.kAEEnd): 'end'
	}
	
	kTypeCompDescriptorOperators = {
			fourcharcode(kae.kAEGreaterThan): 'gt',
			fourcharcode(kae.kAEGreaterThanEquals): 'ge',
			fourcharcode(kae.kAEEquals): 'eq',
			fourcharcode(kae.kAELessThan): 'lt',
			fourcharcode(kae.kAELessThanEquals): 'le',
			fourcharcode(kae.kAEBeginsWith): 'beginswith',
			fourcharcode(kae.kAEEndsWith): 'endswith',
			fourcharcode(kae.kAEContains): 'contains'
	}
	
	kTypeLogicalDescriptorOperators = {
			fourcharcode(kae.kAEAND): 'AND',
			fourcharcode(kae.kAEOR): 'OR',
			fourcharcode(kae.kAENOT): 'NOT'
	}
	
	
	###################################
	
	
	def __init__(self):
		# Clients may add/remove/replace encoder and decoder items:
		self.encoders = {
				AEDesc: self.packdesc,
				type(None): self.packnone,
				bool: self.packbool,
				int: self.packint,
				float: self.packfloat,
				
				bytes: self.packbytes,
				str: self.packstr,
				
				list: self.packlist,
				tuple: self.packlist,
				dict: self.packdict,
				datetime.date: self.packdate,
				datetime.datetime: self.packdatetime,
				datetime.time: self.packtime,
				time.struct_time: self.packstructtime,
				
				mactypes.Alias: self.packalias,
				mactypes.File: self.packfile,
				
				AEType: self.packtype,
				AEEnum: self.packenum,
				AEProp: self.packprop,
				AEKey: self.packkey,
		}
		
		self. decoders = {
				kae.typeNull: self.unpacknull,
				kae.typeBoolean: self.unpackboolean,
				kae.typeFalse: self.unpackfalse,
				kae.typeTrue: self.unpacktrue,
				kae.typeSInt16: self.unpacksint16,
				kae.typeUInt16: self.unpackuint16,
				kae.typeSInt32: self.unpacksint32,
				kae.typeUInt32: self.unpackuint32,
				kae.typeSInt64: self.unpacksint64,
				kae.typeUInt64: self.unpackuint64,
				kae.typeIEEE32BitFloatingPoint: self.unpackfloat32,
				kae.typeIEEE64BitFloatingPoint: self.unpackfloat64,
				kae.type128BitFloatingPoint: self.unpackfloat128,
				
				kae.typeData: self.unpackdata,
				
				kae.typeChar: self.unpackchar,
				kae.typeIntlText: self.unpackintltext,
				kae.typeUTF8Text: self.unpackutf8text,
				kae.typeUTF16ExternalRepresentation: self.unpackutf16externalrepresentation,
				kae.typeStyledText: self.unpackstyledtext,
				kae.typeUnicodeText: self.unpackunicodetext,
				
				kae.typeLongDateTime: self.unpacklongdatetime,
				kae.typeAEList: self.unpackaelist,
				kae.typeAERecord: self.unpackaerecord,
				kae.typeVersion: self.unpackversion,
				
				kae.typeAlias: self.unpackalias,
				kae.typeFSS: self.unpackfss,
				kae.typeFSRef: self.unpackfsref,
				kae.typeFileURL: self.unpackfileurl,
				
				kae.typeQDPoint: self.unpackqdpoint,
				kae.typeQDRectangle: self.unpackqdrect, 
				kae.typeRGBColor: self.unpackrgbcolor,
				
				kae.typeType: self.unpacktype,
				kae.typeEnumeration: self.unpackenumeration,
				kae.typeProperty: self.unpackproperty,
				kae.typeKeyword: self.unpackkeyword,
				
				kae.typeInsertionLoc: self.unpackinsertionloc,
				kae.typeObjectSpecifier: self.unpackobjectspecifier,
				kae.typeAbsoluteOrdinal: self.unpackabsoluteordinal,
				kae.typeCompDescriptor: self.unpackcompdescriptor,
				kae.typeLogicalDescriptor: self.unpacklogicaldescriptor,
				kae.typeRangeDescriptor: self.unpackrangedescriptor,
				
				kae.typeCurrentContainer: lambda desc: self.con,
				kae.typeObjectBeingExamined: lambda desc: self.its,
		}

		self._unittypecodecs = UnitTypeCodecs()
		self._packtextastype = kae.typeUnicodeText
	
	
	###################################
	
	def addunittypes(self, typedefs):
		"""Register custom unit type definitions with this Codecs instance
			e.g. Adobe apps define additional unit types (ciceros, pixels, etc.)
		"""
		self._unittypecodecs.addtypes(typedefs)
	
	def dontcacheunpackedspecifiers(self):
		""" When unpacking object specifiers, unlike AppleScript, appscript caches
			the original AEDesc for efficiency, allowing the resulting reference to
			be re-packed much more quickly. Occasionally this causes compatibility
			problems with applications that returned subtly malformed specifiers.
			To force a Codecs object to fully unpack and repack object specifiers,
			call its dontcacheunpackedspecifiers method.
		"""
		self.unpackobjectspecifier = self.fullyunpackobjectspecifier
	
	def packstringsastype(self, code):
		""" Specify the AE type for packing str objects. Default is kae.typeUnicodeText, but 
			some older non-Unicode-aware Carbon may require kae.typeChar or kae.typeIntlText. 
		"""
		if not (isinstance(code, bytes) and len(code) == 4):
			raise TypeError('Code must be a four-byte value: {!r}'.format(code))
		self._packtextastype = code
	
	
	###################################
	
	def packunknown(self, data):
		"""Clients may override this to provide additional packers."""
		raise TypeError("Can't pack data into an AEDesc (unsupported type): {!r}".format(data))
	
	def unpackunknown(self, desc):
		"""Clients may override this to provide additional unpackers."""
		if desc.isrecord():
			rec = desc.coerce(b'reco')
			rec.setparam(b'pcls', self.pack(AEType(desc.type)))
			decoder = self.decoders.get(b'reco')
			if decoder:
				return decoder(rec)
			else:
				return rec
		else:
			return desc
	
	##
	
	def pack(self, data):
		"""Pack Python data.
			data : anything -- a Python value
			Result : aem.ae.AEDesc -- an Apple event descriptor, or error if no encoder exists for this type of data
		"""
		if isinstance(data, aemreference.Query):
			return data.AEM_packself(self)
		else:
			try:
				return self.encoders[data.__class__](data) # quick lookup by type/class
			except (KeyError, AttributeError):
				for type, encoder in self.encoders.items(): # slower but more thorough lookup that can handle subtypes/subclasses
					if isinstance(data, type):
						return encoder(data)
				didpack, desc = self._unittypecodecs.pack(data)
				if didpack:
					return desc
				else:
					self.packunknown(data)
	
	def unpack(self, desc):
		"""Unpack an Apple event descriptor.
			desc : aem.ae.AEDesc -- an Apple event descriptor
			Result : anything -- a Python value, or the AEDesc object if no decoder is found
		"""
		decoder = self.decoders.get(desc.type)
		if decoder:
			return decoder(desc)
		else:
			didunpack, val = self._unittypecodecs.unpack(desc)
			if didunpack:
				return val
			else:
				return self.unpackunknown(desc)
	
	
	###################################
	
	def packdesc(self, val):
		return val
	
	def packnone(self, val):
		return self.kNullDesc
	
	def packbool(self, val):
		return val and self.kTrueDesc or self.kFalseDesc
	
	def packint(self, val): # note: Python int = C long, so may need to pack as typeSInt64 on 64-bit
		if (-2**31) <= val < (2**31): # pack as typeSInt32 if possible (non-lossy)
			return newdesc(kae.typeSInt32, struct.pack('i', val))
		elif (-2**63) <= val < (2**63): # else pack as typeSInt64 if possible (non-lossy)
			return newdesc(kae.typeSInt64, struct.pack('q', val))
		else: # else pack as typeFloat (lossy)
			return self.pack(float(val))
		
	def packfloat(self, val):
		return newdesc(kae.typeFloat, struct.pack('d', val))
	
	##
	
	def packbytes(self, val):
		return newdesc(kae.typeData, val)
	
	def packstr(self, val):
		# Note: optional BOM is omitted as this causes problems with stupid apps like iTunes 7 that don't
		# handle BOMs correctly; note: while typeUnicodeText is not recommended as of OS 10.4, it's still
		# being used rather than typeUTF8Text or typeUTF16ExternalRepresentation to provide compatibility
		#with not-so-well-designed applications that may have problems with these newer types.
		data = val.encode(nativeutf16encoding)
		if data.startswith(BOM_UTF16_LE) or data.startswith(BOM_UTF16_BE):
			data = data[2:]
		desc = newdesc(kae.typeUnicodeText, data)
		if self._packtextastype == kae.typeUnicodeText:
			return desc
		else:
			return desc.coerce(self._packtextastype)

	##
	
	def packdate(self, val):
		delta = val - self.kShortMacEpoch
		sec = delta.days * 3600 * 24 + delta.seconds
		return newdesc(kae.typeLongDateTime, struct.pack('q', sec))
	
	def packdatetime(self, val):
		delta = val - self.kMacEpoch
		sec = delta.days * 3600 * 24 + delta.seconds
		return newdesc(kae.typeLongDateTime, struct.pack('q', sec))
	
	def packtime(self, val):
		return self.packdatetime(datetime.datetime.combine(datetime.date.today(), val))
	
	def packstructtime(self, val):
		sec = int(time.mktime(val) - self.kMacEpochT)
		return newdesc(kae.typeLongDateTime, struct.pack('q', sec))
	
	def packalias(self, val):
		return val.desc
	packfile = packalias
	
	##
	
	def packlist(self, val):
		lst = newlist()
		for item in val:
			lst.setitem(0, self.pack(item))
		return lst
	
	def packdict(self, val):
		record = newrecord()
		usrf = None
		for key, value in val.items():
			if isinstance(key, (AEType, AEProp)):
				if key.code == b'pcls': # AS packs records that contain a 'class' property by coercing the packed record to that type at the end
					try:
						record = record.coerce(value.code)
					except:
						record.setparam(key.code, self.pack(value))
				else:
					record.setparam(key.code, self.pack(value))
			else:
				if not usrf:
					usrf = newlist()
				usrf.setitem(0, self.pack(key))
				usrf.setitem(0, self.pack(value))
		if usrf:
			record.setparam(b'usrf', usrf)
		return record
	
	##
	
	def packtype(self, val):
		return newdesc(kae.typeType, fourcharcode(val.code))
	
	def packenum(self, val): 
		return newdesc(kae.typeEnumeration, fourcharcode(val.code))
	
	def packprop(self, val): 
		return newdesc(kae.typeProperty, fourcharcode(val.code))
	
	def packkey(self, val): 
		return newdesc(kae.typeKeyword, fourcharcode(val.code))

	
	###################################
	# unpack
	
	def unpacknull(self, desc):
		return None
	
	def unpackboolean(self, desc):
		return bool(desc.data[0])
	
	def unpacktrue(self, desc):
		return True
	
	def unpackfalse(self, desc):
		return False
	
	def unpacksint16(self, desc):
		return struct.unpack('h', desc.data)[0]
	
	def unpackuint16(self, desc):
		return struct.unpack('H', desc.data)[0]
	
	def unpacksint32(self, desc):
		return struct.unpack('i', desc.data)[0]
	
	def unpackuint32(self, desc):
		return struct.unpack('I', desc.data)[0]
	
	def unpacksint64(self, desc):
		return struct.unpack('q', desc.data)[0]
	
	def unpackuint64(self, desc):
		return struct.unpack('Q', desc.data)[0]
	
	def unpackfloat32(self, desc):
		return struct.unpack('f', desc.data)[0]
	
	def unpackfloat64(self, desc):
		return struct.unpack('d', desc.data)[0]
	
	def unpackfloat128(self, desc):
		return struct.unpack('d', desc.coerce(kae.typeIEEE64BitFloatingPoint).data)[0]

	##
	
	def unpackdata(self, desc):
		return desc.data
	
	##
	
	def unpackchar(self, desc):
		return self.unpackunicodetext(desc.coerce(kae.typeUnicodeText))
	
	def unpackintltext(self, desc):
		return self.unpackunicodetext(desc.coerce(kae.typeUnicodeText))
	
	def unpackutf8text(self, desc):
		return desc.data.decode('utf8')
	
	def unpackstyledtext(self, desc):
		return self.unpackunicodetext(desc.coerce(kae.typeUnicodeText))
	
	def unpackunicodetext(self, desc):
		# typeUnicodeText = native endian UTF16 with optional BOM
		data = desc.data
		if data.startswith(BOM_UTF16_BE):
			return data.decode('UTF-16BE')
		elif data.startswith(BOM_UTF16_LE):
			return data.decode('UTF-16LE')
		else:
			return data.decode(nativeutf16encoding)
	
	def unpackutf16externalrepresentation(self, desc): 
		# type UTF16ExternalRepresentation = big-endian UTF16 with optional byte-order-mark 
		# OR little-endian UTF16 with required byte-order-mark
		if desc.data.startswith(BOM_UTF16_LE):
			return desc.data.decode('UTF-16LE')
		else:
			return desc.data.decode('UTF-16BE')
	
	##
	
	def unpacklongdatetime(self, desc):
		return self.kMacEpoch + datetime.timedelta(seconds=struct.unpack('q', desc.data)[0])
	
	def unpackqdpoint(self, desc): 
		x, y = struct.unpack('hh', desc.data)
		return (y, x)
	
	def unpackqdrect(self, desc):
		x1, y1, x2, y2 = struct.unpack('hhhh', desc.data)
		return (y1, x1, y2, x2)
	
	def unpackrgbcolor(self, desc):
		return struct.unpack('HHH', desc.data)
	
	def unpackversion(self, desc):
		# Cocoa apps use unicode strings for version numbers, so return as string for consistency
		try:
			return self.unpack(desc.coerce(kae.typeUnicodeText)) # supported in 10.4+
		except:
			return '{}.{}.{}'.format(desc.data[0], *divmod(desc.data[1], 16)) # note: always big-endian
	
	##
	
	def unpackalias(self, desc):
		return mactypes.Alias.makewithdesc(desc)
		
	def unpackfileurl(self, desc):
		return mactypes.File.makewithdesc(desc)
	unpackfsref = unpackfss = unpackfileurl
	
	##
	
	def unpackaelist(self, desc):
		# Unpack list and its values.
		return [self.unpack(desc.getitem(i + 1, kae.typeWildCard)[1]) for i in range(desc.count())]
	
	def unpackaerecord(self, desc):
		# Unpack record to dict, converting keys from 4-letter codes to AEType instances and unpacking values.
		dct = {}
		for i in range(desc.count()):
			key, value = desc.getitem(i + 1, kae.typeWildCard)
			if key == b'usrf':
				lst = self.unpackaelist(value)
				for i in range(0, len(lst), 2):
					dct[lst[i]] = lst[i+1]
			else:
				dct[AEType(key)] = self.unpack(value)
		return dct

	##
	
	def unpacktype(self, desc):
		return AEType(fourcharcode(desc.data))
	
	def unpackenumeration(self, desc):
		return AEEnum(fourcharcode(desc.data))
	
	def unpackproperty(self, desc):
		return AEProp(fourcharcode(desc.data))
	
	def unpackkeyword(self, desc):
		return AEKey(fourcharcode(desc.data))
					
	##
	
	def fullyunpackobjectspecifier(self, desc):
		# This function performs a full recursive unpacking of object specifiers, reconstructing an 'app'/'con'/'its' based aem reference from the ground up.
		want = self.unpack(desc.getparam(kae.keyAEDesiredClass, kae.typeType)).code # 4-letter code indicating element class
		keyform = self.unpack(desc.getparam(kae.keyAEKeyForm, kae.typeEnumeration)).code # 4-letter code indicating Specifier type
		key = self.unpack(desc.getparam(kae.keyAEKeyData, kae.typeWildCard)) # value indicating which object(s) to select
		ref = self.unpack(desc.getparam(kae.keyAEContainer, kae.typeWildCard)) # recursively unpack container structure
		if not isinstance(ref, aemreference.Query):
			if ref is None:
				ref = self.app
			else:
				ref = self.customroot(ref)
		# print(want, keyform, key, ref) # DEBUG
		if keyform == kae.formPropertyID: # property specifier
			return ref.property(key.code)
		elif keyform == b'usrp': # user-defined property specifier
			return ref.userproperty(key)
		elif keyform == kae.formRelativePosition: # relative element specifier
			if key.code == kae.kAEPrevious:
				return ref.previous(want)
			elif key.code == kae.kAENext:
				return ref.next(want)
			else:
				raise ValueError("Bad relative position selector: {!r}".format(want))
		else: # other element(s) specifier
			ref = ref.elements(want)
			if keyform == kae.formName:
				return ref.byname(key)
			elif keyform == kae.formAbsolutePosition:
				if isinstance(key, _Ordinal):
					if key.code == kae.kAEAll:
						return ref
					else:
						return getattr(ref, {kae.kAEFirst: 'first', kae.kAELast: 'last', kae.kAEMiddle: 'middle', kae.kAEAny: 'any'}[key.code])
				else:
					return ref.byindex(key)
			elif keyform == kae.formUniqueID:
				return ref.byid(key)
			elif keyform == kae.formRange:
				return ref.byrange(*key.range)
			elif keyform == kae.formTest:
				return ref.byfilter(key)
		raise TypeError
	
	
	def unpackobjectspecifier(self, desc):
		# This function performance-optimises the unpacking of some object specifiers by only doing a shallow unpack where only the topmost descriptor is unpacked.
		# The container AEDesc is retained as-is, allowing a full recursive unpack to be performed later on only if needed (e.g. if the __repr__ method is called).
		# For simplicity, only the commonly encountered forms are optimised this way; forms that are rarely returned by applications (e.g. typeRange) are always fully unpacked.
		keyform = self.unpack(desc.getparam(kae.keyAEKeyForm, kae.typeEnumeration)).code
		if keyform in [kae.formPropertyID, kae.formAbsolutePosition, kae.formName, kae.formUniqueID]:
			want = self.unpack(desc.getparam(kae.keyAEDesiredClass, kae.typeType)).code # 4-letter code indicating element class
			key = self.unpack(desc.getparam(kae.keyAEKeyData, kae.typeWildCard)) # value indicating which object(s) to select
			container = aemreference.DeferredSpecifier(desc.getparam(kae.keyAEContainer, kae.typeWildCard), self)
			if keyform == kae.formPropertyID:
				ref = aemreference.Property(want, container, key.code)
			elif keyform == kae.formAbsolutePosition:
				if isinstance(key, _Ordinal):
					if key.code == kae.kAEAll:
						ref = aemreference.AllElements(want, container)
					else:
						keyname = {kae.kAEFirst: 'first', kae.kAELast: 'last', kae.kAEMiddle: 'middle', kae.kAEAny: 'any'}[key.code]
						ref = aemreference.ElementByOrdinal(want, aemreference.UnkeyedElements(want, container), key, keyname)
				else:
					ref = aemreference.ElementByIndex(want, aemreference.UnkeyedElements(want, container), key)
			elif keyform == kae.formName:
				ref = aemreference.ElementByName(want, aemreference.UnkeyedElements(want, container), key)
			elif keyform == kae.formUniqueID:
				ref = aemreference.ElementByID(want, aemreference.UnkeyedElements(want, container), key)
			ref.AEM_packself = lambda codecs:desc
			return ref
		else: # do full unpack of more complex, rarely returned reference forms
			return self.fullyunpackobjectspecifier(desc)
	
	
	def unpackinsertionloc(self, desc):
		return getattr(self.fullyunpackobjectspecifier(desc.getparam(kae.keyAEObject, kae.typeWildCard)), 
				self.kInsertionLocSelectors[desc.getparam(kae.keyAEPosition, kae.typeEnumeration).data])
	
	
	def unpackcompdescriptor(self, desc):
		operator = self.kTypeCompDescriptorOperators[desc.getparam(kae.keyAECompOperator, kae.typeEnumeration).data]
		op1 = self.unpack(desc.getparam(kae.keyAEObject1, kae.typeWildCard))
		op2 = self.unpack(desc.getparam(kae.keyAEObject2, kae.typeWildCard))
		if operator == 'contains':
			if isinstance(op1, aemreference.Query) and op1.AEM_root() == aemreference.its:
				return op1.contains(op2)
			else:
				return op2.isin(op1)
		return getattr(op1, operator)(op2)
	
	
	def unpacklogicaldescriptor(self, desc):
		operator = self.kTypeLogicalDescriptorOperators[desc.getparam(kae.keyAELogicalOperator, kae.typeEnumeration).data]
		operands = self.unpack(desc.getparam(kae.keyAELogicalTerms, kae.typeAEList))
		return operator == 'NOT' and operands[0].NOT or getattr(operands[0], operator)(*operands[1:])
	
	def unpackrangedescriptor(self, desc):
		return _Range([self.unpack(desc.getparam(kae.keyAERangeStart, kae.typeWildCard)), 
				self.unpack(desc.getparam(kae.keyAERangeStop, kae.typeWildCard))])
	
	
	def unpackabsoluteordinal(self, desc):
		return _Ordinal(fourcharcode(desc.data))
	
	##
	
	app = aemreference.app
	con = aemreference.con
	its = aemreference.its
	customroot = aemreference.customroot
	




"""connect - Creates Apple event descriptor records of typeProcessSerialNumber and typeApplicationURL, used to specify the target application in send.Event() constructor. """
import struct
from time import sleep

from . import ae, kae

from .aemsend import Event

__all__ = ['launchapp', 'processexistsforpath', 'processexistsforpid', 'processexistsforurl', 
		'processexistsfordesc', 'currentapp', 'localapp', 'remoteapp', 'CantLaunchApplicationError']

######################################################################
# PRIVATE
######################################################################

kLSLaunchDefaults = 0x00000001
kLSLaunchAndPrint = 0x00000002
kLSLaunchReserved2 = 0x00000004
kLSLaunchReserved3 = 0x00000008
kLSLaunchReserved4 = 0x00000010
kLSLaunchReserved5 = 0x00000020
kLSLaunchAndDisplayErrors = 0x00000040
kLSLaunchInhibitBGOnly = 0x00000080
kLSLaunchDontAddToRecents = 0x00000100
kLSLaunchDontSwitch = 0x00000200
kLSLaunchNoParams = 0x00000800
kLSLaunchAsync = 0x00010000
kLSLaunchStartClassic = 0x00020000
kLSLaunchInClassic = 0x00040000
kLSLaunchNewInstance = 0x00080000
kLSLaunchAndHide = 0x00100000
kLSLaunchAndHideOthers = 0x00200000
kLSLaunchHasUntrustedContents = 0x00400000


_kNoProcess = 0
_kCurrentProcess = 2

_nulladdressdesc = ae.newdesc(kae.typeProcessSerialNumber, struct.pack('II', 0, _kNoProcess)) # ae.newappleevent complains if you pass None as address, so we give it one to throw away

_launchevent = Event(_nulladdressdesc, b'ascrnoop').AEM_event
_runevent = Event(_nulladdressdesc, b'aevtoapp').AEM_event

#######

def _launchapplication(path, event, newinstance=False, hide=False):
	flags = kLSLaunchNoParams | kLSLaunchStartClassic | kLSLaunchDontSwitch
	if newinstance:
		flags |= kLSLaunchNewInstance
	if hide:
		flags |= kLSLaunchAndHide
	try:
		return ae.launchapplication(path, event, flags)
	except ae.MacOSError as err:
		raise CantLaunchApplicationError(err.args[0], path) from err


######################################################################
# PUBLIC
######################################################################


class CantLaunchApplicationError(Exception):
	
	_lserrors = {
		# following taken from <http://developer.apple.com/documentation/Carbon/Reference/LaunchServicesReference>:
		-10660: "The application cannot be run because it is inside a Trash folder.",
		-10810: "An unknown error has occurred.",
		-10811: "The item to be registered is not an application.",
		-10813: "Data of the desired type is not available (for example, there is no kind string).",
		-10814: "No application in the Launch Services database matches the input criteria.",
		-10817: "Data is structured improperly (for example, an item's information property list is malformed).",
		-10818: "A launch of the application is already in progress.",
		-10822: "There is a problem communicating with the server process that maintains the Launch Services database.",
		-10823: "The filename extension to be hidden cannot be hidden.",
		-10825: "The application to be launched cannot run on the current Mac OS version.",
		-10826: "The user does not have permission to launch the application (on a managed network).",
		-10827: "The executable file is missing or has an unusable format.",
		-10828: "The Classic emulation environment was required but is not available.",
		-10829: "The application to be launched cannot run simultaneously in two different user sessions.",
	}

	def __init__(self, errornumber, apppath):
		self._number = errornumber
		self._apppath = apppath
		Exception.__init__(self, errornumber, apppath)
	
	errornumber = property(lambda self: self._number, doc="int -- Mac OS error number")
	
	apppath = property(lambda self: self._apppath, doc="str -- application path")
	
	def __int__(self):
		return self._number
	
	def __str__(self):
		return "Can't launch application at {!r}: {} ({})".format(self._apppath, self._lserrors.get(self._number, 'OS error'), self._number)


def launchapp(path, newinstance=False, hide=False):
	"""Send a 'launch' event to an application. If application is not already running, it will be launched in background first.
		path : string -- full path to application, e.g. '/Applications/TextEdit.app'
		newinstance : bool -- launch a new application instance?
		hide : bool -- hide after launch?
		Result : AEAddressDesc
	"""
	if newinstance:
		desc = _launchapplication(path, _launchevent, newinstance, hide)
	else:
		try:
			# If app is already running, calling ae.launchapplication will send a 'reopen' event, so need to check for this first:
			desc = ae.psnforapplicationpath(path)
		except ae.MacOSError as err:
			if err.args[0] == -600: # Application isn't running, so launch it and send it a 'launch' event:
				sleep(1)
				desc = _launchapplication(path, _launchevent, newinstance, hide)
			else:
				raise
		else: # App is already running, so send it a 'launch' event:
			ae.newappleevent(b'ascr', b'noop', desc, kae.kAutoGenerateReturnID, 
					kae.kAnyTransactionID).send(kae.kAEWaitReply, kae.kAEDefaultTimeout)
	return desc

##

def processexistsforpath(path):
	"""Does a local process launched from the specified application file exist?
		Note: if path is invalid, a MacOSError is raised.
	"""
	try:
		ae.psnforapplicationpath(path)
		return True
	except ae.MacOSError as err:
		if err.args[0] == -600: 
			return False
		else:
			raise

def processexistsforpid(pid):
	"""Is there a local application process with the given unix process id?"""
	return bool(ae.isvalidpid(pid))

def processexistsforurl(url):
	"""Does an application process specified by the given eppc:// URL exist?
		Note: this will send a 'launch' Apple event to the target application.
	"""
	if ':' not in url: # workaround: process will crash if no colon in URL (OS bug)
		raise ValueError("Invalid url: {!r}".format(url))
	return processexistsfordesc(ae.newdesc(kae.typeApplicationURL, url))

def processexistsfordesc(desc):
	"""Does an application process specified by the given AEAddressDesc exist?
		Returns false if process doesn't exist OR remote Apple events aren't allowed.
		Note: this will send a 'launch' Apple event to the target application.
	"""
	try:
		# This will usually raise error -1708 if process is running, and various errors
		# if the process doesn't exist/can't be reached. If app is running but busy,
		# AESendMessage() may return a timeout error (this should be -1712, but
		# -609 is often returned instead for some reason).
		Event(desc, b'ascrnoop').send()
	except ae.MacOSError as err:
		return err.args[0] not in [-600, -905] # -600 = no process; -905 = no network access
	return True


#######

currentapp = ae.newdesc(kae.typeProcessSerialNumber, struct.pack('II', 0, _kCurrentProcess))


def localapp(path, newinstance=False, hide=False):
	"""Make an AEAddressDesc identifying a local application. (Application will be launched if not already running.)
		path : string -- full path to application, e.g. '/Applications/TextEdit.app'
		newinstance : bool -- launch a new application instance?
		hide : bool -- hide after launch?
		Result : AEAddressDesc
	"""
	# Always create AEAddressDesc by process serial number; that way there's no confusion if multiple versions of the same app are running
	if newinstance:
		desc = _launchapplication(path, _runevent, newinstance, hide)
	else:
		try:
			desc = ae.psnforapplicationpath(path)
		except ae.MacOSError as err:
			if err.args[0] == -600: # Application isn't running, so launch it in background and send it a standard 'run' event.
				sleep(1)
				desc = _launchapplication(path, _runevent, newinstance, hide)
			else:
				raise
	return desc


def localappbypid(pid):
	"""Make an AEAddressDesc identifying a local process.
		pid : integer -- Unix process id
		Result : AEAddressDesc
	"""
	return ae.newdesc(kae.typeKernelProcessID, struct.pack('i', pid))


def remoteapp(url):
	"""Make an AEAddressDesc identifying a running application on another machine.
		url : string -- URL for remote application, e.g. 'eppc://user:password@192.168.2.1/TextEdit'
		Result : AEAddressDesc
	"""
	if ':' not in url: # workaround: process will crash if no colon in URL (OS bug)
		raise ValueError("Invalid url: {!r}".format(url))
	return ae.newdesc(kae.typeApplicationURL, url)




"""genericreference -- allows user to construct relative (con- and its- based) references without immediate access to application terminology. """

import aem

######################################################################
# PUBLIC
######################################################################

class GenericReference(object):
	def __init__(self, call):
		self._call = call # list of recorded call information (inital value is ['app'], ['con'] or ['its'])
	
	def __getattr__(self, i):
		return GenericReference(self._call + [('__getattr__', i, '.{}')])
	
	def __getitem__(self, i):
		return GenericReference(self._call + [('__getitem__', i, '[{!r}]')])
	
	def __call__(self, *args, **kargs):
		return GenericReference(self._call + [('__call__', (args, kargs), None)])
	
	def __gt__(self, i):
		return GenericReference(self._call + [('AS__gt__', i, ' > {!r}')])
	
	def __ge__(self, i):
		return GenericReference(self._call + [('AS__ge__', i, ' >= {!r}')])
	
	def __eq__(self, i):
		return GenericReference(self._call + [('AS__eq__', i, ' == {!r}')])
	
	def __ne__(self, i):
		return GenericReference(self._call + [('AS__ne__', i, ' != {!r}')])
	
	def __lt__(self, i):
		return GenericReference(self._call + [('AS__lt__', i, ' < {!r}')])
	
	def __le__(self, i):
		return GenericReference(self._call + [('AS__le__', i, ' <= {!r}')])
	
	def __hash__(self):
		return hash(self._call)
		
	def __repr__(self):
		s = self._call[0]
		for method, args, repstr in self._call[1:]:
			if method == '__call__':
				s += '({})'.format(', '.join(['{!r}'.format(i) for i in args[0]] + ['{}={!r}'.format(i) for i in args[1].items()]))
			elif method == '__getitem__' and isinstance(args, slice):
				s+= '[{!r}:{!r}]'.format(args.start, args.stop)
			elif method == '__getattr__' and args in ['AND', 'OR', 'NOT']:
				s = '({}).{}'.format(s, args)
			else:
				s += repstr.format(args)
		return s
	
	def AS_resolve(self, Reference, appdata):
		# (Note: reference.Reference class is passed as argument simply to avoid circular import between that module and this)
		ref = Reference(appdata, {'app':aem.app, 'con':aem.con, 'its':aem.its}[self._call[0]])
		for method, args, repstr in self._call[1:]:
			if method == '__getattr__':
				ref = getattr(ref, args)
			elif method == '__call__':
				ref = ref(*args[0], **args[1])
			else:
				ref = getattr(ref, method)(args)
		return ref

##

con = GenericReference(['con'])
its = GenericReference(['its'])
# 'app' is defined in reference module




"""reservedkeywords """

import keyword

# Important: the following must be reserved:
#
# - names of properties and methods used in reference.Application and reference.Reference classes
# - names of built-in keyword arguments in reference.Command.__call__

kReservedKeywords = [
	"ID",
	"beginning",
	"end",
	"before",
	"after",
	"previous",
	"next",
	"first",
	"middle",
	"last",
	"any",
	"beginswith",
	"endswith",
	"contains",
	"isin",
	"doesnotbeginwith",
	"doesnotendwith",
	"doesnotcontain",
	"isnotin",
	"AND",
	"NOT",
	"OR",
	"begintransaction",
	"aborttransaction",
	"endtransaction",
	"isrunning",
	"permissiontoautomate", # 10.14+
	"resulttype",
	"canaskforconsent", # 10.14+
	"ignore",
	"timeout",
	"waitreply",
	"help",
	"relaunchmode",
	"as",
	"with",
	"True",
	"False",
	"None",
	 ] + keyword.kwlist


"""terminology -- Obtains an application's aete resource(s) using a 'ascrgdte' event and converts them into lookup tables for use in AppData objects. """

from aem import Application, AEType, AEEnum, EventError, findapp, ae, kae
from . import defaultterminology

from .terminologyparser import buildtablesforaetes
from .keywordwrapper import Keyword

__all__ = ['tablesforapp', 'tablesformodule', 'tablesforaetes', 
		'kProperty', 'kElement', 'kCommand', 
		'defaulttables', 'aetesforapp', 'dump']


######################################################################
# PUBLIC
######################################################################
# Constants

kProperty = b'p'
kElement = b'e'
kCommand = b'c'

######################################################################
# PRIVATE
######################################################################
# Cache

_terminologycache = {} # cache parsed terminology

######################################################################
# Default terminology tables for converting between human-readable identifiers and Apple event codes; used by all apps.
# Includes default entries for Required Suite, get/set and miscellaneous other commands; application may override some or all of these definitions.

# Type tables; used to translate constants
# e.g. k.document <-> AEType('docu')
# e.g. k.ask <-> AEEnum('ask ')

_defaulttypebyname = {} # used to encode class and enumerator keywords
_defaulttypebycode = {} # used to decode class (typeType) and enumerator (typeEnum) descriptors
_defaulttypecodebyname = {} # used to check for name collisions

for _, enumerators in defaultterminology.enumerations:
	for name, code in enumerators:
		_defaulttypebyname[name] = AEEnum(code)
		_defaulttypebycode[code] = Keyword(name)
		_defaulttypecodebyname[name] = code
for defs in [defaultterminology.types, defaultterminology.properties]:
	for name, code in defs:
		_defaulttypebyname[name] = AEType(code)
		_defaulttypebycode[code] = Keyword(name)
		_defaulttypecodebyname[name] = code

# Reference tables; used to translate references and commands
# e.g. app(...).documents.text <-> app.elements('docu').property('ctxt')
# e.g. app(...).quit(saving=k.ask) <-> Application(...).event('aevtquit', {'savo': AEEnum('ask ')})

_defaultreferencebycode = {} # used to decode property and element specifiers
_defaultreferencebyname = {} # used to encode property and element specifiers and Apple events
_defaultcommandcodebyname = {} # used to check for name collisions


for name, code in defaultterminology.properties:
	_defaultreferencebycode[kProperty + code] = (kProperty, name)
	_defaultreferencebyname[name] = (kProperty, code)

for name, code in defaultterminology.elements:
	_defaultreferencebycode[kElement + code] = (kElement, name)
	_defaultreferencebyname[name] = (kElement, code)

for name, code, params in defaultterminology.commands:
	_defaultreferencebyname[name] = (kCommand, (code, dict(params)))
	_defaultcommandcodebyname[name] = code


######################################################################
# Translation table parsers

def _maketypetable(classes, enums, properties):
	# Used for constructing k.keywords
	# Each argument is of format [[name, code], ...]
	typebycode = _defaulttypebycode.copy()
	typebyname = _defaulttypebyname.copy()
	# note: testing indicates that where name+code clashes occur, classes have highest priority, followed by properties, with enums last (prior to 0.19.0 this code gave higher priority to enums):
	for klass, table in [(AEEnum, enums), (AEType, properties), (AEType, classes)]: # note: packing properties as AEProp causes problems when the same name is used for both a class and a property, and the property's definition masks the class's one (e.g. Finder's 'file'); if an AEProp is passed where an AEType is expected, it can cause an error as it's not what the receiving app expects. (Whereas they may be more tolerant of an AEType being passed where an AEProp is expected.) Also, note that AppleScript always seems to pack property names as typeType, so we should be ok following its lead here.
		for i, (name, code) in enumerate(table):
			# If an application-defined name overlaps an existing type name but has a different code, append '_' to avoid collision:
			if _defaulttypecodebyname.get(name, code) != code:
				name += '_'
			typebycode[code] = Keyword(name) # to handle synonyms, if same code appears more than once then use name from last definition in list
			name, code = table[-i - 1]
			if _defaulttypecodebyname.get(name, code) != code:
				name += '_'
			typebyname[name] = klass(code) # to handle synonyms, if same name appears more than once then use code from first definition in list
	return typebycode, typebyname


def _makereferencetable(properties, elements, commands):
	# Used for constructing references and commands
	# First two parameters are of format [[name, code], ...]
	# Last parameter is of format [name, code, direct arg type, [[arg code, arg name], ...]]
	referencebycode = _defaultreferencebycode.copy()
	referencebyname = _defaultreferencebyname.copy()
	for kind, table in [(kElement, elements), (kProperty, properties)]:
		# note: if property and element names are same (e.g. 'file' in BBEdit), will pack as property specifier unless it's a special case (i.e. see 'text' below). Note that there is currently no way to override this, i.e. to force appscript to pack it as an all-elements specifier instead (in AS, this would be done by prepending the 'every' keyword), so clients would need to use aem for that (but could add an 'all' method to Reference class if there was demand for a built-in workaround)
		for i, (name, code) in enumerate(table):
			# If an application-defined name overlaps an existing type name but has a different code, append '_' to avoid collision:
			if _defaulttypecodebyname.get(name, code) != code:
				name += '_'
			referencebycode[kind+code] = (kind, name) # to handle synonyms, if same code appears more than once then use name from last definition in list
			name, code = table[-i - 1]
			if _defaulttypecodebyname.get(name, code) != code:
				name += '_'
			referencebyname[name] = (kind, code) # to handle synonyms, if same name appears more than once then use code from first definition in list
	if 'text' in referencebyname: # special case: AppleScript always packs 'text of...' as all-elements specifier
		referencebyname['text'] = (kElement, referencebyname['text'][1])
	for name, code, args in commands[::-1]: # to handle synonyms, if two commands have same name but different codes, only the first definition should be used (iterating over the commands list in reverse ensures this)
		# Avoid collisions between default commands and application-defined commands with same name but different code (e.g. 'get' and 'set' in InDesign CS2):
		if code != _defaultcommandcodebyname.get(name, code):
			name += '_'
		referencebyname[name] = (kCommand, (code, dict(args)))
	return referencebycode, referencebyname


######################################################################
# PUBLIC
######################################################################


defaulttables = _maketypetable([], [], []) + _makereferencetable([], [], []) # (typebycode, typebyname, referencebycode, referencebyname)


def aetesforapp(aemapp):
	"""Get aetes from local/remote app via an ascrgdte event; result is a list of byte strings."""
	try:
		aetes = aemapp.event(b'ascrgdte', {b'----':0}).send(120 * 60)
	except Exception as e: # (e.g.application not running)
		if isinstance(e, EventError) and e.errornumber == -192:
			aetes = []
		else:
			raise RuntimeError("Can't get terminology for application ({!r}): {}".format(aemapp, e)) from e
	if not isinstance(aetes, list):
		aetes = [aetes]
	return [aete for aete in aetes if isinstance(aete, ae.AEDesc) and aete.type == kae.typeAETE and aete.data]


def tablesforaetes(aetes):
	"""Build terminology tables from a list of unpacked aete byte strings.
		Result : tuple of dict -- (typebycode, typebyname, referencebycode, referencebyname)
	"""
	classes, enums, properties, elements, commands = buildtablesforaetes(aetes)
	return _maketypetable(classes, enums, properties) + _makereferencetable(properties, elements, commands)


def tablesformodule(terms):
	"""Build terminology tables from a dumped terminology module.
		Result : tuple of dict -- (typebycode, typebyname, referencebycode, referencebyname)
	"""
	return _maketypetable(terms.classes, terms.enums, terms.properties) \
			+ _makereferencetable(terms.properties, terms.elements, terms.commands)


def tablesforapp(aemapp):
	"""Build terminology tables for an application.
		aemapp : aem.Application
		Result : tuple of dict -- (typebycode, typebyname, referencebycode, referencebyname)
	"""
	if aemapp.AEM_identity not in _terminologycache:
		_terminologycache[aemapp.AEM_identity] = tablesforaetes(aetesforapp(aemapp))
	return _terminologycache[aemapp.AEM_identity]


def dumptables(tables, sourcepath, modulepath):
	"""Dump terminology data to Python module.
		tables : tuple of list -- five-item tuple: (classes, enums, properties, elements, commands)
		sourcepath : str -- path to source application/scripting addition
		modulepath : str -- path to generated module
	"""
	from pprint import pprint
	atts = zip(('classes', 'enums', 'properties', 'elements', 'commands'), tables)
	f = open(modulepath, 'w')
	f.write('version = 1.1\n')
	f.write('path = {!r}\n'.format(sourcepath))
	for key, value in atts:
		if key[0] != '_':
			f.write('\n{} = \\\n'.format(key))
			pprint(value, f)
	f.close()


######################################################################
# PUBLIC
######################################################################


def dump(apppath, modulepath):
	"""Dump application terminology data to Python module.
		apppath : str -- name or path of application
		modulepath : str -- path to generated module
		
	Generates a Python module containing an application's basic terminology 
	(names and codes) as used by appscript.
	
	Call the dump() function to dump faulty aetes to Python module, e.g.:
	
		dump('MyApp', '/path/to/site-packages/myappglue.py')
	
	Patch any errors by hand, then import the patched module into your script 
	and pass it to appscript's app() constructor via its 'terms' argument, e.g.:
	
		from appscript import *
		import myappglue
		
		myapp = app('MyApp', terms=myappglue)

	Note that dumped terminologies aren't used by appscript's built-in help system.
	"""
	apppath = findapp.byname(apppath)
	tables = buildtablesforaetes(aetesforapp(Application(apppath)))
	dumptables(tables, apppath, modulepath)
